import typing
import enum
from abc import ABC, abstractmethod
from collections import namedtuple

BERObject = namedtuple('BERObject', ['tag', 'content'])

class IncompleteBERError(ValueError):
	def __init__(self, expected_length=-1):
		super().__init__()
		self.expected_length = expected_length

def decode_ber(data):
	index = 0
	if len(data) < 2:
		raise IncompleteBERError(2)
	identifier = data[index]
	ber_class = identifier >> 6
	ber_constructed = bool(identifier & 0x20)
	ber_type = identifier & 0x1f
	index += 1
	if not data[index] & 0x80:
		length = data[index]
		index += 1
	elif data[index] == 0x80:
		raise ValueError('Indefinite form not implemented')
	elif data[index] == 0xff:
		raise ValueError('BER length invalid')
	else:
		num = data[index] & ~0x80
		index += 1
		if len(data) < index + num:
			raise IncompleteBERError(index + num)
		length = 0
		for octet in data[index:index + num]:
			length = length << 8 | octet
		index += num
	if len(data) < index + length:
		raise IncompleteBERError(index + length)
	ber_content = data[index: index + length]
	rest = data[index + length:]
	return BERObject((ber_class, ber_constructed, ber_type), ber_content), rest

def encode_ber(obj):
	tag = (obj.tag[0] & 0b11) << 6 | (obj.tag[1] & 1) << 5 | (obj.tag[2] & 0b11111)
	length = len(obj.content)
	if length < 127:
		return bytes([tag, length]) + obj.content
	octets = []
	while length:
		octets.append(length & 0xff)
		length = length >> 8
	return bytes([tag, 0x80 | len(octets)]) + bytes(reversed(octets)) + obj.content

class BERType(ABC):
	@classmethod
	@abstractmethod
	def from_ber(cls, data):
		raise NotImplementedError()

	@classmethod
	@abstractmethod
	def to_ber(cls, obj):
		raise NotImplementedError()

	def __bytes__(self):
		return type(self).to_ber(self)

class OctetString(BERType):
	BER_TAG = (0, False, 4)

	@classmethod
	def from_ber(cls, data):
		obj, rest = decode_ber(data)
		if obj.tag != cls.BER_TAG:
			raise ValueError('Expected tag %s but found %s'%(cls.BER_TAG, obj.tag))
		return obj.content, rest

	@classmethod
	def to_ber(cls, obj):
		if not isinstance(obj, bytes):
			raise TypeError()
		return encode_ber(BERObject(cls.BER_TAG, obj))

class Integer(BERType):
	BER_TAG = (0, False, 2)

	@classmethod
	def from_ber(cls, data):
		obj, rest = decode_ber(data)
		if obj.tag != cls.BER_TAG:
			raise ValueError()
		return int.from_bytes(obj.content, 'big', signed=True), rest

	@classmethod
	def to_ber(cls, obj):
		if not isinstance(obj, int):
			raise TypeError()
		if obj < 0:
			res = obj.to_bytes((8 + (obj + 1).bit_length()) // 8, byteorder='big', signed=True)
		else:
			res = obj.to_bytes(max(1, (obj.bit_length() + 7) // 8), 'big', signed=False)
			if res[0] & 0x80:
				res = b'\x00' + res
		return encode_ber(BERObject(cls.BER_TAG, res))

class Boolean(BERType):
	BER_TAG = (0, False, 1)

	@classmethod
	def from_ber(cls, data):
		obj, rest = decode_ber(data)
		if obj.tag != cls.BER_TAG or len(obj.content) != 1:
			raise ValueError()
		return bool(obj.content[0]), rest

	@classmethod
	def to_ber(cls, obj):
		if not isinstance(obj, bool):
			raise TypeError()
		content = b'\xff' if obj else b'\x00'
		return encode_ber(BERObject(cls.BER_TAG, content))

class Set(BERType):
	BER_TAG = (0, True, 17)
	SET_TYPE: BERType

	@classmethod
	def from_ber(cls, data):
		setobj, rest = decode_ber(data)
		if setobj.tag != cls.BER_TAG:
			raise ValueError()
		objs = []
		data = setobj.content
		while data:
			obj, data = cls.SET_TYPE.from_ber(data)
			objs.append(obj)
		return list(objs), rest

	@classmethod
	def to_ber(cls, obj):
		content = b''
		for item in obj:
			content += cls.SET_TYPE.to_ber(item)
		return encode_ber(BERObject(cls.BER_TAG, content))

class SequenceOf(Set):
	BER_TAG = (0, True, 16)

class Sequence(BERType):
	BER_TAG = (0, True, 16)
	# (Type, attr_name, default_value, optional?)
	SEQUENCE_FIELDS: typing.ClassVar[typing.List[typing.Tuple[BERType, str, typing.Any, bool]]] = []

	def __init__(self, *args, **kwargs):
		for index, spec in enumerate(type(self).SEQUENCE_FIELDS):
			# pylint: disable=consider-using-get,unused-variable
			field_type, name, default, optional = spec
			if index < len(args):
				value = args[index]
			elif name in kwargs:
				value = kwargs[name]
			else:
				value = default() if callable(default) else default
			setattr(self, name, value)

	def __repr__(self):
		args = []
			# pylint: disable=unused-variable
		for field_type, name, default, optional in type(self).SEQUENCE_FIELDS:
			args.append('%s=%s'%(name, repr(getattr(self, name))))
		return '<%s(%s)>'%(type(self).__name__, ', '.join(args))

	@classmethod
	def from_ber(cls, data):
		seqobj, rest = decode_ber(data)
		if seqobj.tag != cls.BER_TAG:
			raise ValueError()
		args = []
		data = seqobj.content
		# pylint: disable=unused-variable
		for field_type, name, default, optional in cls.SEQUENCE_FIELDS:
			try:
				obj, data = field_type.from_ber(data)
				args.append(obj)
			except ValueError as e:
				if not optional:
					raise e
				args.append(None)
		return cls(*args), rest

	@classmethod
	def to_ber(cls, obj):
		if not isinstance(obj, cls):
			raise TypeError()
		content = b''
		# pylint: disable=unused-variable
		for field_type, name, default, optional in cls.SEQUENCE_FIELDS:
			if not optional or getattr(obj, name) is not None:
				content += field_type.to_ber(getattr(obj, name))
		return encode_ber(BERObject(cls.BER_TAG, content))

class Choice(BERType):
	BER_TAG: typing.ClassVar[typing.Tuple[int, bool, int]]

	@classmethod
	def from_ber(cls, data):
		obj, rest = decode_ber(data)
		for subcls in cls.__subclasses__():
			if subcls.BER_TAG == obj.tag:
				return subcls.from_ber(data)
		return None, rest

	@classmethod
	def to_ber(cls, obj):
		for subcls in cls.__subclasses__():
			if isinstance(obj, subcls):
				return subcls.to_ber(obj)
		raise TypeError()

class Wrapper(BERType):
	BER_TAG: typing.ClassVar[typing.Tuple[int, bool, int]]
	WRAPPED_ATTRIBUTE: typing.ClassVar[str]
	WRAPPED_TYPE: typing.ClassVar[BERType]
	WRAPPED_DEFAULT: typing.ClassVar[typing.Any]
	WRAPPED_CLSATTRS: typing.ClassVar[typing.Dict[str, typing.Any]] = {}

	def __init__(self, *args, **kwargs):
		cls = type(self)
		attribute = cls.WRAPPED_ATTRIBUTE
		if args:
			setattr(self, attribute, args[0])
		elif kwargs:
			setattr(self, attribute, kwargs[attribute])
		else:
			setattr(self, attribute, cls.WRAPPED_DEFAULT() if callable(cls.WRAPPED_DEFAULT) else cls.WRAPPED_DEFAULT)

	def __repr__(self):
		return '<%s(%s)>'%(type(self).__name__, repr(getattr(self, type(self).WRAPPED_ATTRIBUTE)))

	@classmethod
	def from_ber(cls, data):
		class WrappedType(cls.WRAPPED_TYPE):
			BER_TAG = cls.BER_TAG
		for key, value in cls.WRAPPED_CLSATTRS.items():
			setattr(WrappedType, key, value)
		value, rest = WrappedType.from_ber(data)
		return cls(value), rest

	@classmethod
	def to_ber(cls, obj):
		class WrappedType(cls.WRAPPED_TYPE):
			BER_TAG = cls.BER_TAG
		for key, value in cls.WRAPPED_CLSATTRS.items():
			setattr(WrappedType, key, value)
		if not isinstance(obj, cls):
			raise TypeError()
		return WrappedType.to_ber(getattr(obj, cls.WRAPPED_ATTRIBUTE))

def retag(cls, tag):
	class Overwritten(cls):
		BER_TAG = tag
	return Overwritten

class Enum(Integer):
	BER_TAG = (0, False, 10)
	ENUM_TYPE = typing.ClassVar[enum.Enum]

	@classmethod
	def from_ber(cls, data):
		value, rest = super().from_ber(data)
		return cls.ENUM_TYPE(value), rest

	@classmethod
	def to_ber(cls, obj):
		if not isinstance(obj, cls.ENUM_TYPE):
			raise TypeError()
		return super().to_ber(obj.value)

def wrapenum(enumtype):
	class WrappedEnum(Enum):
		ENUM_TYPE = enumtype
	return WrappedEnum



import traceback
import ssl
import socketserver
import typing
import logging
import time
import random
import string
import itertools

from . import asn1, exceptions, ldap, schema, objects

def pop_control(controls, oid):
	result = None
	remaining_controls = []
	for control in controls or []:
		if control.controlType == oid:
			result = control
			break
		remaining_controls.append(control)
	return result, remaining_controls

def reject_critical_controls(controls=None):
	for control in controls or []:
		if control.criticality:
			raise exceptions.LDAPUnavailableCriticalExtension()

def mark_last(iterable):
	'''Yield (item, is_last) for all items in iterable

	is_last is True for the last items and False for other items.'''
	prev_item = None
	for item in iterable:
		if prev_item is not None:
			yield prev_item, False
		prev_item = item
	if prev_item is not None:
		yield prev_item, True

class RequestLogAdapter(logging.LoggerAdapter):
	def process(self, msg, kwargs):
		return self.extra['trace_id'] + ': ' + msg, kwargs

class BaseLDAPRequestHandler(socketserver.BaseRequestHandler):
	logger = logging.getLogger('ldapserver.server')

	def setup(self):
		super().setup()
		self.trace_id = ''.join([random.choice(string.ascii_letters) for _ in range(10)])
		self.logger = RequestLogAdapter(self.logger, {'trace_id': self.trace_id})
		self.keep_running = True

	def handle(self):
		time_connect = time.perf_counter()
		self.logger.info('Connection from %r', self.client_address)
		buf = b''
		while self.keep_running:
			try:
				shallowmsg, buf = ldap.ShallowLDAPMessage.from_ber(buf)
				for respmsg in self.handle_message(shallowmsg):
					self.request.sendall(ldap.LDAPMessage.to_ber(respmsg))
			except asn1.IncompleteBERError:
				chunk = self.request.recv(4096)
				if not chunk:
					self.keep_running = False
					self.request.close()
				else:
					buf += chunk
		self.request.close()
		time_disconnect = time.perf_counter()
		self.logger.info('Disconnected duration_seconds=%.3f', time_disconnect - time_connect)

	def handle_message(self, shallowmsg: ldap.ShallowLDAPMessage) -> typing.Iterable[ldap.LDAPMessage]:
		'''Handle an LDAP request foobar

		:param shallowmsg: Half-decoded LDAP message to handle
		:returns: Response messages
		'''
		msgtypes = {
			ldap.BindRequest: (self.handle_bind, ldap.BindResponse),
			ldap.UnbindRequest: (self.handle_unbind, None),
			ldap.SearchRequest: (self.handle_search, ldap.SearchResultDone),
			ldap.ModifyRequest: (self.handle_modify, ldap.ModifyResponse),
			ldap.AddRequest: (self.handle_add,  ldap.AddResponse),
			ldap.DelRequest: (self.handle_delete, ldap.DelResponse),
			ldap.ModifyDNRequest: (self.handle_modifydn, ldap.ModifyDNResponse),
			ldap.CompareRequest: (self.handle_compare, ldap.CompareResponse),
			ldap.AbandonRequest: (self.handle_abandon, None),
			ldap.ExtendedRequest: (self.handle_extended, ldap.ExtendedResponse),
		}
		handler, response_type = msgtypes.get(shallowmsg.protocolOpType, (None, None))
		try:
			if handler is None:
				raise exceptions.LDAPProtocolError()
			try:
				msg = shallowmsg.decode()[0]
			except ValueError as e:
				self.logger.error('Could not decode message %s, ignoring', shallowmsg)
				raise exceptions.LDAPProtocolError() from e
			for args in handler(msg.protocolOp, msg.controls):
				response, controls = args if isinstance(args, tuple) else (args, None)
				yield ldap.LDAPMessage(shallowmsg.messageID, response, controls)
		except exceptions.LDAPError as e:
			if response_type is not None:
				respmsg = ldap.LDAPMessage(shallowmsg.messageID, response_type(e.code, diagnosticMessage=e.message))
				yield respmsg
				self.logger.info('Operation aborted, responded with result code "%s" msg="%s"', e.code.name, e.message)
		except Exception as e: # pylint: disable=broad-except
			if response_type is not None:
				respmsg = ldap.LDAPMessage(shallowmsg.messageID, response_type(ldap.LDAPResultCode.other))
				yield respmsg
				self.logger.exception('Uncaught exception, responded with result code "other"')
			else:
				self.logger.exception('Uncaught exception, ignored request')

	def handle_bind(self, op: ldap.BindRequest, controls=None) -> typing.Iterable[ldap.ProtocolOp]:
		self.logger.info('BIND %s', op)
		reject_critical_controls(controls)
		raise exceptions.LDAPAuthMethodNotSupported()

	def handle_unbind(self, op: ldap.UnbindRequest, controls=None) -> typing.NoReturn:
		self.logger.info('UNBIND %s', op)
		reject_critical_controls(controls)
		self.keep_running = False
		return []

	def handle_search(self, op: ldap.SearchRequest, controls=None) -> typing.Iterable[ldap.ProtocolOp]:
		self.logger.info('SEARCH %s', op)
		reject_critical_controls(controls)
		yield ldap.SearchResultDone(ldap.LDAPResultCode.success)

	def handle_modify(self, op: ldap.ModifyRequest, controls=None) -> typing.Iterable[ldap.ProtocolOp]:
		self.logger.info('MODIFY %s', op)
		reject_critical_controls(controls)
		raise exceptions.LDAPInsufficientAccessRights()

	def handle_add(self, op: ldap.AddRequest, controls=None) -> typing.Iterable[ldap.ProtocolOp]:
		self.logger.info('ADD %s', op)
		reject_critical_controls(controls)
		raise exceptions.LDAPInsufficientAccessRights()

	def handle_delete(self, op: ldap.DelRequest, controls=None) -> typing.Iterable[ldap.ProtocolOp]:
		self.logger.info('DELETE %s', op)
		reject_critical_controls(controls)
		raise exceptions.LDAPInsufficientAccessRights()

	def handle_modifydn(self, op: ldap.ModifyDNRequest, controls=None) -> typing.Iterable[ldap.ProtocolOp]:
		self.logger.info('MODIFYDN %s', op)
		reject_critical_controls(controls)
		raise exceptions.LDAPInsufficientAccessRights()

	def handle_compare(self, op: ldap.CompareRequest, controls=None) -> typing.Iterable[ldap.ProtocolOp]:
		self.logger.info('COMPRAE %s', op)
		reject_critical_controls(controls)
		raise exceptions.LDAPInsufficientAccessRights()

	def handle_abandon(self, op: ldap.AbandonRequest, controls=None) -> typing.NoReturn:
		self.logger.info('ABANDON %s', op)
		reject_critical_controls(controls)

	def handle_extended(self, op: ldap.ExtendedRequest, controls=None) -> typing.Iterable[ldap.ProtocolOp]:
		self.logger.info('EXTENDED %s', op)
		reject_critical_controls(controls)
		raise exceptions.LDAPProtocolError()

class LDAPRequestHandler(BaseLDAPRequestHandler):
	'''
	.. py:attribute:: rootdse

		Special :any:`LDAPObject` that contains information
		about the server, such as supported extentions and SASL authentication
		mechansims. Attributes can be accessed in a dict-like fashion.
	'''

	subschema = objects.SubschemaSubentry(schema.RFC4519_SCHEMA, 'cn=Subschema', cn=['Subschema'])
	'''
	.. py:attribute:: subschema

		Special :any:`LDAPObject` that describes the schema.
		Per default the subschema includes standard syntaxes, standard matching
		rules and objectclasses/attributetypes for the rootdse and subschema.
		It does not include objectclasses/attributetypes for actual data
		(e.g. users and groups). See :any:`Subschema` for details.
	'''

	static_objects = tuple()

	def setup(self):
		super().setup()
		self.rootdse = self.subschema.RootDSE()
		self.rootdse['objectClass'] = ['top']
		if self.supports_starttls:
			self.rootdse['supportedExtension'].append(ldap.STARTTLS_OID)
		if self.supports_whoami:
			self.rootdse['supportedExtension'].append(ldap.WHOAMI_OID)
		if self.supports_password_modify:
			self.rootdse['supportedExtension'].append(ldap.PASSWORD_MODIFY_OID)
		if self.supports_paged_results:
			self.rootdse['supportedControl'].append(ldap.PAGED_RESULTS_OID)
		if self.supports_sasl_anonymous:
			self.rootdse['supportedSASLMechanisms'].append('ANONYMOUS')
		if self.supports_sasl_plain:
			self.rootdse['supportedSASLMechanisms'].append('PLAIN')
		if self.supports_sasl_external:
			self.rootdse['supportedSASLMechanisms'].append('EXTERNAL')
		self.rootdse['supportedLDAPVersion'] = ['3']
		self.bind_object = None
		self.bind_sasl_state = None
		self.__paged_searches = {} # pagination cookie -> (iterator, orig_op)
		self.__paged_cookie_counter = 0

	def handle_bind(self, op, controls=None):
		reject_critical_controls(controls)
		if op.version != 3:
			raise exceptions.LDAPProtocolError('Unsupported protocol version')
		auth = op.authentication
		# Resume ongoing SASL dialog
		if self.bind_sasl_state and isinstance(auth, ldap.SaslCredentials) \
				and auth.mechanism == self.bind_sasl_state[0]:
			mechanism, iterator = self.bind_sasl_state
			self.bind_sasl_state = None
			resp_code = ldap.LDAPResultCode.saslBindInProgress
			try:
				resp = iterator.send(auth.credentials)
				self.bind_sasl_state = (mechanism, iterator)
			except StopIteration as e:
				resp_code = ldap.LDAPResultCode.success
				self.bind_object, resp = e.value # pylint: disable=unpacking-non-sequence
			yield ldap.BindResponse(resp_code, serverSaslCreds=resp)
			return
		# If auth type or SASL method changed, abort SASL dialog
		self.bind_sasl_state = None
		if isinstance(auth, ldap.SimpleAuthentication):
			self.bind_object = self.do_bind_simple(op.name, auth.password)
			yield ldap.BindResponse(ldap.LDAPResultCode.success)
		elif isinstance(auth, ldap.SaslCredentials):
			ret = self.do_bind_sasl(auth.mechanism, auth.credentials)
			if isinstance(ret, tuple):
				self.bind_object, resp = ret
				yield ldap.BindResponse(ldap.LDAPResultCode.success, serverSaslCreds=resp)
				return
			iterator = iter(ret)
			resp_code = ldap.LDAPResultCode.saslBindInProgress
			try:
				resp = next(iterator)
				self.bind_sasl_state = (auth.mechanism, iterator)
			except StopIteration as e:
				resp_code = ldap.LDAPResultCode.success
				self.bind_object, resp = e.value # pylint: disable=unpacking-non-sequence
			yield ldap.BindResponse(resp_code, serverSaslCreds=resp)
		else:
			yield from super().handle_bind(op, controls) # pylint: disable=not-an-iterable

	def do_bind_simple(self, dn='', password=b''):
		'''Do LDAP BIND with simple authentication

		:param dn: Distinguished name of object to be authenticated or empty
		:type dn: str
		:param password: Password, may be empty
		:type password: bytes

		:returns: Bind object
		:rtype: obj

		Delegates implementation to :any:`do_bind_simple_anonymous`,
		:any:`do_bind_simple_unauthenticated` or :any:`do_bind_simple_authenticated`
		according to `RFC 4513`_.'''
		if not dn and not password:
			return self.do_bind_simple_anonymous()
		if not password:
			return self.do_bind_simple_unauthenticated(dn)
		return self.do_bind_simple_authenticated(dn, password)

	def do_bind_simple_anonymous(self):
		'''Do LDAP BIND with simple anonymous authentication (`RFC 4513 5.1.1.`_)

		:raises exceptions.LDAPError: if authentication failed

		:returns: Bind object on success
		:rtype: obj

		Calld by :any:`do_bind_simple`. Always returns None.'''
		return None

	def do_bind_simple_unauthenticated(self, dn):
		'''Do LDAP BIND with simple unauthenticated authentication (`RFC 4513 5.1.2.`_)

		:param dn: Distinguished name of the object to be authenticated
		:type dn: str

		:raises exceptions.LDAPError: if authentication failed

		:returns: Bind object on success
		:rtype: obj

		Calld by :any:`do_bind_simple`. The default implementation always raises an
		:any:`LDAPInvalidCredentials` exception.'''
		raise exceptions.LDAPInvalidCredentials()

	def do_bind_simple_authenticated(self, dn, password):
		'''Do LDAP BIND with simple name/password authentication (`RFC 4513 5.1.3.`_)

		:param dn: Distinguished name of the object to be authenticated
		:type dn: str
		:param password: Password for object
		:type dn: bytes

		:raises exceptions.LDAPError: if authentication failed

		:returns: Bind object on success
		:rtype: obj

		Calld by :any:`do_bind_simple`. The default implementation always raises an
		`LDAPInvalidCredentials` exception.'''
		raise exceptions.LDAPInvalidCredentials()

	def do_bind_sasl(self, mechanism, credentials=None, dn=None):
		'''Do LDAP BIND with SASL authentication (RFC 4513 and 4422)

		:param mechanism: Name of the selected SASL mechanism
		:type mechanism: str
		:param credentials: Initial client response
		:type credentials: bytes, optional
		:param dn: Distinguished name in LDAP BIND request, should be ignored for
		           SASL authentication
		:type dn: str, optional

		:returns: Bind object and final server challenge, only returns on success
		:rtype: Tuple (obj, bytes/None)

		The call only returns if authentication succeeded. In any other case,
		an appropriate :any:`exceptions.LDAPError` is raised.

		Some SASL methods require additional challenge-response round trips. These
		can be achieved with the `yield` statement:

		    client_response = yield server_challenge

		Generally all server challenges and client responses can always be absent
		(indicated by None), empty (empty bytes object) or consist of any number
		of bytes. Whether a challenge or response may or must be absent or present
		is defined by the individual SASL mechanism.

		IANA list of SASL mechansims: https://www.iana.org/assignments/sasl-mechanisms/sasl-mechanisms.xhtml
		'''
		if not mechanism:
			# Request to abort current negotiation (RFC4513 5.2.1.2)
			raise exceptions.LDAPAuthMethodNotSupported()
		if mechanism == 'ANONYMOUS' and self.supports_sasl_anonymous:
			if credentials is not None:
				credentials = credentials.decode()
			return self.do_bind_sasl_anonymous(trace_info=credentials), None
		if mechanism == 'PLAIN' and self.supports_sasl_plain:
			if credentials is None:
				raise exceptions.LDAPProtocolError('Unsupported protocol version')
			authzid, authcid, password = credentials.split(b'\0', 2)
			return self.do_bind_sasl_plain(authcid.decode(), password.decode(), authzid.decode() or None), None
		if mechanism == 'EXTERNAL' and self.supports_sasl_external:
			if credentials is not None:
				credentials = credentials.decode()
			return self.do_bind_sasl_external(authzid=credentials), None
		raise exceptions.LDAPAuthMethodNotSupported()

	supports_sasl_anonymous = False

	def do_bind_sasl_anonymous(self, trace_info=None):
		'''Do LDAP BIND with SASL "ANONYMOUS" mechanism (RFC 4505)

		:param trace_info: Trace information, either an email address or an
		                   opaque string that does not contain the '@' character
		:type trace_info: str, optional

		:raises exceptions.LDAPError: if authentication failed

		:returns: Bind object on success
		:rtype: obj

		Calld by :any:`do_bind_sasl`. The default implementation raises an
		:any:`LDAPAuthMethodNotSupported` exception.'''
		raise exceptions.LDAPAuthMethodNotSupported()

	supports_sasl_plain = False

	def do_bind_sasl_plain(self, identity, password, authzid=None):
		'''Do LDAP BIND with SASL "PLAIN" mechanism (RFC 4616)

		:param identity: Authentication identity (authcid)
		:type identity: str
		:param password: Password (passwd)
		:type password: str
		:param authzid: Authorization identity
		:type authzid: str, optional

		:raises exceptions.LDAPError: if authentication failed

		:returns: Bind object on success
		:rtype: obj

		Calld by :any:`do_bind_sasl`. The default implementation raises an
		:any:`LDAPAuthMethodNotSupported` exception.'''
		raise exceptions.LDAPAuthMethodNotSupported()

	supports_sasl_external = False

	def do_bind_sasl_external(self, authzid=None):
		'''Do LDAP BIND with SASL "EXTERNAL" mechanism (RFC 4422 and 4513)

		:param authzid: Authorization identity
		:type authzid: str, optional

		:raises exceptions.LDAPError: if authentication failed

		:returns: Bind object on success
		:rtype: obj

		EXTERNAL is commonly used for TLS client certificate authentication or
		system user based authentication on UNIX sockets.

		Calld by :any:`do_bind_sasl`. The default implementation raises an
		:any:`LDAPAuthMethodNotSupported` exception.'''
		raise exceptions.LDAPAuthMethodNotSupported()

	supports_paged_results = True

	def __handle_search_paged(self, op, paged_control, controls=None):
		def build_control(size=0, cookie=b''):
			value = ldap.PagedResultsValue(size=size, cookie=cookie)
			return ldap.Control(controlType=ldap.PAGED_RESULTS_OID,
			                    criticality=True, controlValue=bytes(value))

		# pylint: disable=no-member
		paged_control = ldap.PagedResultsValue.from_ber(paged_control.controlValue)[0]
		if not paged_control.cookie: # New paged search request
			results = self.do_search(op.baseObject, op.scope, op.filter)
			results = map(lambda obj: obj.search(op.baseObject, op.scope, op.filter, op.attributes, op.typesOnly), results)
			results = filter(None, results)
			iterator = iter(mark_last(results))
		else: # Continue existing paged search
			try:
				iterator, orig_op = self.__paged_searches.pop(paged_control.cookie)
			except KeyError as exc:
				raise exceptions.LDAPUnwillingToPerform('Invalid pagination cookie') from exc
			if ldap.ProtocolOp.to_ber(orig_op) != ldap.ProtocolOp.to_ber(op):
				raise exceptions.LDAPUnwillingToPerform('Search parameter mismatch')
			if not paged_control.size: # Cancel paged search
				yield ldap.SearchResultDone(ldap.LDAPResultCode.success), [build_control()]
				return
		is_last = True
		entries = 0
		time_start = time.perf_counter()
		for entry, is_last in itertools.islice(iterator, 0, paged_control.size):
			self.logger.debug('SEARCH entry %r', entry)
			entries += 1
			yield entry
		cookie = b''
		if not is_last:
			cookie = str(self.__paged_cookie_counter).encode()
			self.__paged_cookie_counter += 1
			self.__paged_searches[cookie] = iterator, op
		yield ldap.SearchResultDone(ldap.LDAPResultCode.success), [build_control(cookie=cookie)]
		time_end = time.perf_counter()
		self.logger.info('SEARCH dn=%r dn_scope=%s filter=%s attributes=%r page_cookie=%r entries=%d duration_seconds=%.3f',
		                 op.baseObject, op.scope.name, op.filter, ' '.join(op.attributes), cookie, entries, time_end - time_start)

	def handle_search(self, op, controls=None):
		self.logger.debug('SEARCH request dn=%r dn_scope=%s filter=%r attributes=%r',
		                  op.baseObject, op.scope.name, str(op.filter), ' '.join(op.attributes))
		paged_control = None
		if self.supports_paged_results:
			paged_control, controls = pop_control(controls, ldap.PAGED_RESULTS_OID)
		reject_critical_controls(controls)
		if paged_control:
			yield from self.__handle_search_paged(op, paged_control, controls)
			return
		entries = 0
		time_start = time.perf_counter()
		for obj in self.do_search(op.baseObject, op.scope, op.filter):
			entry = obj.search(op.baseObject, op.scope, op.filter, op.attributes, op.typesOnly)
			if entry:
				self.logger.debug('SEARCH entry %r', entry)
				entries += 1
				yield entry
		yield ldap.SearchResultDone(ldap.LDAPResultCode.success)
		time_end = time.perf_counter()
		self.logger.info('SEARCH dn=%r dn_scope=%s filter=\'%s\' attributes=%r entries=%d duration_seconds=%.3f',
		                 op.baseObject, op.scope.name, op.filter, ' '.join(op.attributes), entries, time_end - time_start)

	def do_search(self, baseobj, scope, filterobj):
		'''Do LDAP SEARCH operation

		:param baseobj: Distinguished name of the LDAP entry relative to which the
		                search is to be performed
		:type baseobj: str
		:param scope: Search scope
		:type scope: SearchScope
		:param filterobj: Filter object
		:type filterobj: Filter

		:raises exceptions.LDAPError: on error

		:returns: Iterable of dn, attributes tuples

		The default implementation returns matching objects from the root dse and
		the subschema.'''
		yield self.rootdse
		yield self.subschema
		yield from self.static_objects

	def handle_compare(self, op, controls=None):
		self.logger.info('COMPRAE request "%s" %s=%s', op.entry, op.ava.attributeDesc, repr(op.ava.assertionValue))
		obj = self.do_compare(op.entry, op.ava.attributeDesc, op.ava.assertionValue)
		if obj is None:
			raise exceptions.LDAPNoSuchObject()
		if obj.match_compare(op.ava.attributeDesc, op.ava.assertionValue):
			self.logger.info('COMPRAE result TRUE')
			return [ldap.CompareResponse(ldap.LDAPResultCode.compareTrue)]
		else:
			self.logger.info('COMPRAE result FALSE')
			return [ldap.CompareResponse(ldap.LDAPResultCode.compareFalse)]

	def do_compare(self, dn, attribute, value):
		'''Lookup object for COMPARE operation

		:param dn: Distinguished name of the LDAP entry
		:type dn: str
		:param attribute: Attribute type
		:type attribute: str
		:param value: Attribute value
		:type value: bytes

		:raises exceptions.LDAPError: on error

		:returns: `Object` or None

		The default implementation calls `do_search` and returns the first object
		with the right DN.'''
		objs = self.do_search(dn, ldap.SearchScope.baseObject, ldap.FilterPresent(attribute='objectClass'))
		for obj in objs:
			try:
				obj.compare(dn)
				return obj
			except exceptions.LDAPNoSuchObject:
				pass
			except exceptions.LDAPError:
				return obj
		raise exceptions.LDAPNoSuchObject()

	def handle_unbind(self, op, controls=None):
		self.logger.info('UNBIND')
		reject_critical_controls(controls)
		self.keep_running = False
		return []

	def handle_extended(self, op, controls=None):
		reject_critical_controls(controls)
		if op.requestName == ldap.STARTTLS_OID and self.supports_starttls:
			self.logger.info('EXTENDED STARTTLS')
			# StartTLS (RFC 4511)
			yield ldap.ExtendedResponse(ldap.LDAPResultCode.success, responseName=ldap.STARTTLS_OID)
			try:
				self.do_starttls()
			except Exception: # pylint: disable=broad-except
				traceback.print_exc()
				self.keep_running = False
			if ldap.STARTTLS_OID in self.rootdse['supportedExtension']:
				self.rootdse['supportedExtension'].remove(ldap.STARTTLS_OID)
		elif op.requestName == ldap.WHOAMI_OID and self.supports_whoami:
			self.logger.info('EXTENDED WHOAMI')
			# "Who am I?" Operation (RFC 4532)
			identity = (self.do_whoami() or '').encode()
			yield ldap.ExtendedResponse(ldap.LDAPResultCode.success, responseValue=identity)
		elif op.requestName == ldap.PASSWORD_MODIFY_OID and self.supports_password_modify:
			self.logger.info('EXTENDED PASSWORD_MODIFY')
			# Password Modify Extended Operation (RFC 3062)
			newpw = None
			if op.requestValue is None:
				newpw = self.do_password_modify()
			else:
				decoded, _ = ldap.PasswdModifyRequestValue.from_ber(op.requestValue)
				# pylint: disable=no-member
				newpw = self.do_password_modify(decoded.userIdentity, decoded.oldPasswd, decoded.newPasswd)
			if newpw is None:
				yield ldap.ExtendedResponse(ldap.LDAPResultCode.success)
			else:
				encoded = ldap.PasswdModifyResponseValue.to_ber(ldap.PasswdModifyResponseValue(newpw))
				yield ldap.ExtendedResponse(ldap.LDAPResultCode.success, responseValue=encoded)
		else:
			self.logger.warning('Unsupported or disabled EXTENDED operation %r', op.requestName)
			yield from super().handle_extended(op, controls) # pylint: disable=not-an-iterable

	#: :any:`ssl.SSLContext` for StartTLS
	ssl_context = None

	@property
	def supports_starttls(self):
		'''
		'''
		return self.ssl_context is not None and not isinstance(self.request, ssl.SSLSocket)

	def do_starttls(self):
		'''Do StartTLS extended operation (RFC 4511)

		Called by `handle_extended()` if :any:`supports_starttls` is True. The default
		implementation uses `ssl_context`.

		Note that the (success) response to the request is sent before this method
		is called. If a call to this method fails, the LDAP connection is
		immediately terminated.'''
		self.request = self.ssl_context.wrap_socket(self.request, server_side=True)

	#:
	supports_whoami = False

	def do_whoami(self):
		'''Do "Who am I?" extended operation (RFC 4532)

		:returns: Current authorization identity (authzid) or empty string for anonymous sessions
		:rtype: str

		Called by `handle_extended()` if `supports_whoami` is True. The default
		implementation always returns an empty string.'''
		return ''

	#:
	supports_password_modify = False

	def do_password_modify(self, user=None, old_password=None, new_password=None):
		'''Do password modify extended operation (RFC 3062)

		:param user: User the request relates to, may or may not be a
		             distinguished name. If absent, the request relates to the
		             user currently associated with the LDAP connection
		:type user: str, optional
		:param old_password: Current password of user
		:type old_password: bytes, optional
		:param new_password: Desired password for user
		:type new_password: bytes, optional

		Called by `handle_extended()` if :any:`supports_password_modify` is True. The
		default implementation always raises an :any:`LDAPUnwillingToPerform` error.'''
		raise exceptions.LDAPUnwillingToPerform()



import typing
from abc import ABC, abstractmethod
import enum

from . import asn1

class LDAPString(asn1.OctetString):
	@classmethod
	def from_ber(cls, data):
		raw, rest = super().from_ber(data)
		return raw.decode(), rest

	@classmethod
	def to_ber(cls, obj):
		if not isinstance(obj, str):
			raise TypeError()
		return super().to_ber(obj.encode())

class LDAPOID(LDAPString):
	pass

class AttributeValueAssertion(asn1.Sequence):
	SEQUENCE_FIELDS = [
		(LDAPString, 'attributeDesc', None, False),
		(asn1.OctetString, 'assertionValue', None, False),
	]

	attributeDesc: str
	assertionValue: bytes

def escape_filter_assertionvalue(value):
	allowed_bytes = b'0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!"#$%&\'+,-./:;<=>?@[\\]^_`{|}~ '
	res = []
	for byte in value:
		if byte not in allowed_bytes:
			res += b'\\%02X'%byte
		else:
			res.append(byte)
	return bytes(res).decode()

class Filter(asn1.Choice, ABC):
	'''Base class for filters in SEARCH operations'''

	@abstractmethod
	def __str__(self):
		raise NotImplementedError()

class FilterAnd(asn1.Wrapper, Filter):
	'''AND conjunction of multiple filters ``(&filters...)``

	.. py:attribute:: filters
		:type: list
		:value: []

		List of :any:`Filter` objects
	'''
	BER_TAG = (2, True, 0)
	WRAPPED_ATTRIBUTE = 'filters'
	WRAPPED_TYPE = asn1.Set
	WRAPPED_CLSATTRS = {'SET_TYPE': Filter}

	filters: typing.List[Filter]

	def __init__(self, filters=None):
		super().__init__(filters=filters)

	def __str__(self):
		return '(&%s)'%(''.join([str(subfilter) for subfilter in self.filters]))

class FilterOr(asn1.Wrapper, Filter):
	'''OR conjunction of multiple filters ``(|filters...)``

	.. py:attribute:: filters
		:type: list
		:value: []

		List of :any:`Filter` objects
	'''
	BER_TAG = (2, True, 1)
	WRAPPED_ATTRIBUTE = 'filters'
	WRAPPED_TYPE = asn1.Set
	WRAPPED_CLSATTRS = {'SET_TYPE': Filter}

	filters: typing.List[Filter]

	def __init__(self, filters=None):
		super().__init__(filters=filters)

	def __str__(self):
		return '(|%s)'%(''.join([str(subfilter) for subfilter in self.filters]))

class FilterNot(asn1.Sequence, Filter):
	'''Negation of a filter ``(!filter)``

	.. py:attribute:: filter
		:type: Filter
	'''

	BER_TAG = (2, True, 2)
	SEQUENCE_FIELDS = [
		(Filter, 'filter', None, False)
	]

	filter: Filter

	# pylint: disable=redefined-builtin
	def __init__(self, filter=None):
		super().__init__(filter=filter)

	def __str__(self):
		return '(!%s)'%str(self.filter)

class FilterEqual(asn1.Sequence, Filter):
	'''Attribute equal filter ``(attribute=value)``

	.. py:attribute:: attribute
		:type: str
	.. py:attribute:: value
		:type: bytes
	'''
	BER_TAG = (2, True, 3)
	SEQUENCE_FIELDS = [
		(LDAPString, 'attribute', None, False),
		(asn1.OctetString, 'value', None, False)
	]

	attribute: str
	value: bytes

	def __init__(self, attribute=None, value=None):
		super().__init__(attribute=attribute, value=value)

	def __str__(self):
		return '(%s=%s)'%(self.attribute, escape_filter_assertionvalue(self.value))

class Substring(asn1.Choice, ABC):
	pass

class InitialSubstring(asn1.Wrapper, Substring):
	BER_TAG = (2, False, 0)

	WRAPPED_ATTRIBUTE = 'value'
	WRAPPED_TYPE = asn1.OctetString
	WRAPPED_DEFAULT = b''

class AnySubstring(asn1.Wrapper, Substring):
	BER_TAG = (2, False, 1)

	WRAPPED_ATTRIBUTE = 'value'
	WRAPPED_TYPE = asn1.OctetString
	WRAPPED_DEFAULT = b''

class FinalSubstring(asn1.Wrapper, Substring):
	BER_TAG = (2, False, 2)

	WRAPPED_ATTRIBUTE = 'value'
	WRAPPED_TYPE = asn1.OctetString
	WRAPPED_DEFAULT = b''

class Substrings(asn1.SequenceOf):
	SET_TYPE = Substring

class FilterSubstrings(asn1.Sequence, Filter):
	'''Attribute substrings filter ``(attribute=initial*any*final)``

	.. py:attribute:: attribute
		:type: str
	.. py:attribute:: inital
		:type: bytes
	.. py:attribute:: any
		:type: bytes
	'''
	BER_TAG = (2, True, 4)
	SEQUENCE_FIELDS = [
		(LDAPString, 'attribute', None, False),
		(Substrings, 'substrings', lambda: [], False)
	]

	attribute: str
	substrings: typing.List[Substring]

	def __init__(self, attribute=None, substrings=None):
		super().__init__(attribute=attribute, substrings=substrings)

	@property
	def initial_substring(self):
		results = [substring.value for substring in self.substrings if isinstance(substring, InitialSubstring)]
		if len(results) != 1:
			return None
		return results[0]

	@property
	def any_substrings(self):
		return [substring.value for substring in self.substrings if isinstance(substring, AnySubstring)]

	@property
	def final_substring(self):
		results = [substring.value for substring in self.substrings if isinstance(substring, FinalSubstring)]
		if len(results) != 1:
			return None
		return results[0]

	def __str__(self):
		substrings = [self.initial_substring or b''] + self.any_substrings + [self.final_substring or b'']
		value = '*'.join(map(escape_filter_assertionvalue, substrings))
		return f'({self.attribute}={value})'

class FilterGreaterOrEqual(asn1.Sequence, Filter):
	BER_TAG = (2, True, 5)
	SEQUENCE_FIELDS = [
		(LDAPString, 'attribute', None, False),
		(asn1.OctetString, 'value', None, False)
	]

	attribute: str
	value: bytes

	def __init__(self, attribute=None, value=None):
		super().__init__(attribute=attribute, value=value)

	def __str__(self):
		return '(%s>=%s)'%(self.attribute, escape_filter_assertionvalue(self.value))

class FilterLessOrEqual(asn1.Sequence, Filter):
	BER_TAG = (2, True, 6)
	SEQUENCE_FIELDS = [
		(LDAPString, 'attribute', None, False),
		(asn1.OctetString, 'value', None, False)
	]

	attribute: str
	value: bytes

	def __init__(self, attribute=None, value=None):
		super().__init__(attribute=attribute, value=value)

	def __str__(self):
		return '(%s<=%s)'%(self.attribute, escape_filter_assertionvalue(self.value))

class FilterPresent(asn1.Wrapper, Filter):
	'''Attribute present filter ``(attribute=*)``

	.. py:attribute:: attribute
		:type: str
	'''
	BER_TAG = (2, False, 7)
	WRAPPED_ATTRIBUTE = 'attribute'
	WRAPPED_TYPE = LDAPString
	WRAPPED_DEFAULT = None

	attribute: str

	def __init__(self, attribute=None):
		super().__init__(attribute=attribute)

	def __str__(self):
		return '(%s=*)'%(self.attribute)

class FilterApproxMatch(asn1.Sequence, Filter):
	BER_TAG = (2, True, 8)
	SEQUENCE_FIELDS = [
		(LDAPString, 'attribute', None, False),
		(asn1.OctetString, 'value', None, False)
	]

	attribute: str
	value: bytes

	def __init__(self, attribute=None, value=None):
		super().__init__(attribute=attribute, value=value)

	def __str__(self):
		return '(%s~=%s)'%(self.attribute, escape_filter_assertionvalue(self.value))

class FilterExtensibleMatch(asn1.Sequence, Filter):
	BER_TAG = (2, True, 9)
	SEQUENCE_FIELDS = [
		(asn1.retag(LDAPString, (2, False, 1)), 'matchingRule', None, True),
		(asn1.retag(LDAPString, (2, False, 2)), 'type', None, True),
		(asn1.retag(asn1.OctetString, (2, False, 3)), 'matchValue', None, False),
		(asn1.retag(asn1.Boolean, (2, False, 4)), 'dnAttributes', None, True),
	]

	matchingRule: str
	type: str
	matchValue: bytes
	dnAttributes: bool

	def __str__(self):
		key = ''
		if self.type is not None:
			key += self.type
		if self.dnAttributes:
			key += ':dn'
		if self.matchingRule is not None:
			key += ':' + self.matchingRule
		return '(%s:=%s)'%(key, escape_filter_assertionvalue(self.matchValue))

class SearchScope(enum.Enum):
	''':any:`enum.Enum` of `scope` values in SEARCH operations'''
	# pylint: disable=invalid-name
	#: Search is constrained to the base object
	baseObject = 0
	#: Search is constrained to the immediate subordinates of the base object
	singleLevel = 1
	#: Search is constrained to the base object and to all its subordinates
	wholeSubtree = 2

class DerefAliases(enum.Enum):
	# pylint: disable=invalid-name
	neverDerefAliases = 0
	derefInSearching = 1
	derefFindingBaseObj = 2
	derefAlways = 3

class LDAPResultCode(enum.Enum):
	# pylint: disable=invalid-name
	success                      = 0
	operationsError              = 1
	protocolError                = 2
	timeLimitExceeded            = 3
	sizeLimitExceeded            = 4
	compareFalse                 = 5
	compareTrue                  = 6
	authMethodNotSupported       = 7
	strongerAuthRequired         = 8
	# -- 9 reserved --
	referral                     = 10
	adminLimitExceeded           = 11
	unavailableCriticalExtension = 12
	confidentialityRequired      = 13
	saslBindInProgress           = 14
	noSuchAttribute              = 16
	undefinedAttributeType       = 17
	inappropriateMatching        = 18
	constraintViolation          = 19
	attributeOrValueExists       = 20
	invalidAttributeSyntax       = 21
	# -- 22-31 unused --
	noSuchObject                 = 32
	aliasProblem                 = 33
	invalidDNSyntax              = 34
	# -- 35 reserved for undefined isLeaf --
	aliasDereferencingProblem    = 36
	# -- 37-47 unused --
	inappropriateAuthentication  = 48
	invalidCredentials           = 49
	insufficientAccessRights     = 50
	busy                         = 51
	unavailable                  = 52
	unwillingToPerform           = 53
	loopDetect                   = 54
	# -- 55-63 unused --
	namingViolation              = 64
	objectClassViolation         = 65
	notAllowedOnNonLeaf          = 66
	notAllowedOnRDN              = 67
	entryAlreadyExists           = 68
	objectClassModsProhibited    = 69
	# -- 70 reserved for CLDAP --
	affectsMultipleDSAs          = 71
	# -- 72-79 unused --
	other                        = 80

class LDAPResult(asn1.Sequence):
	BER_TAG = (5, True, 1)
	SEQUENCE_FIELDS = [
		(asn1.wrapenum(LDAPResultCode), 'resultCode', LDAPResultCode.success, False),
		(LDAPString, 'matchedDN', '', False),
		(LDAPString, 'diagnosticMessage', '', False),
	]

	resultCode: LDAPResultCode
	matchedDN: str
	diagnosticMessage: str

class AttributeSelection(asn1.SequenceOf):
	SET_TYPE = LDAPString

class AuthenticationChoice(asn1.Choice):
	pass

class SimpleAuthentication(asn1.Wrapper, AuthenticationChoice):
	BER_TAG = (2, False, 0)
	WRAPPED_ATTRIBUTE = 'password'
	WRAPPED_TYPE = asn1.OctetString
	WRAPPED_DEFAULT = b''

	password: bytes

	def __repr__(self):
		if not self.password:
			return '<%s(EMPTY PASSWORD)>'%(type(self).__name__)
		return '<%s(PASSWORD HIDDEN)>'%(type(self).__name__)

class SaslCredentials(asn1.Sequence, AuthenticationChoice):
	BER_TAG = (2, True, 3)
	SEQUENCE_FIELDS = [
		(LDAPString, 'mechanism', None, False),
		(asn1.OctetString, 'credentials', None, True),
	]

	mechanism: str
	credentials: bytes

class AttributeValueSet(asn1.Set):
	SET_TYPE = asn1.OctetString

class PartialAttribute(asn1.Sequence):
	SEQUENCE_FIELDS = [
		(LDAPString, 'type', None, False),
		(AttributeValueSet, 'vals', lambda: [], False),
	]

	type: str
	vals: typing.List[bytes]

class PartialAttributeList(asn1.SequenceOf):
	SET_TYPE = PartialAttribute

class Attribute(asn1.Sequence):
	# Constrain: vals must not be empty
	SEQUENCE_FIELDS = [
		(LDAPString, 'type', None, False),
		(AttributeValueSet, 'vals', lambda: [], False),
	]

	type: str
	vals: typing.List[bytes]

class AttributeList(asn1.SequenceOf):
	SET_TYPE = Attribute

class ProtocolOp(asn1.Choice):
	pass

class BindRequest(asn1.Sequence, ProtocolOp):
	BER_TAG = (1, True, 0)
	SEQUENCE_FIELDS = [
		(asn1.Integer, 'version', 3, False),
		(LDAPString, 'name', '', False),
		(AuthenticationChoice, 'authentication', lambda: SimpleAuthentication(), False) # pylint: disable=unnecessary-lambda
	]

	version: int
	name: str
	authentication: AuthenticationChoice

class BindResponse(asn1.Sequence, ProtocolOp):
	BER_TAG = (1, True, 1)
	SEQUENCE_FIELDS = [
		(asn1.wrapenum(LDAPResultCode), 'resultCode', None, False),
		(LDAPString, 'matchedDN', '', False),
		(LDAPString, 'diagnosticMessage', '', False),
		(asn1.retag(asn1.OctetString, (2, False, 7)), 'serverSaslCreds', None, True)
	]

	resultCode: LDAPResultCode
	matchedDN: str
	diagnosticMessage: str
	serverSaslCreds: bytes

class UnbindRequest(asn1.Sequence, ProtocolOp):
	BER_TAG = (1, False, 2)

class SearchRequest(asn1.Sequence, ProtocolOp):
	BER_TAG = (1, True, 3)
	SEQUENCE_FIELDS = [
		(LDAPString, 'baseObject', '', False),
		(asn1.wrapenum(SearchScope), 'scope', SearchScope.wholeSubtree, False),
		(asn1.wrapenum(DerefAliases), 'derefAliases', DerefAliases.neverDerefAliases, False),
		(asn1.Integer, 'sizeLimit', 0, False),
		(asn1.Integer, 'timeLimit', 0, False),
		(asn1.Boolean, 'typesOnly', False, False),
		(Filter, 'filter', lambda: FilterPresent('objectClass'), False),
		(AttributeSelection, 'attributes', lambda: [], False)
	]

	baseObject: str
	scope: SearchScope
	derefAliases: DerefAliases
	sizeLimit: int
	timeLimit: int
	typesOnly: bool
	filter: Filter
	attributes: typing.List[str]

	@classmethod
	def from_ber(cls, data):
		return super().from_ber(data)

class SearchResultEntry(asn1.Sequence, ProtocolOp):
	BER_TAG = (1, True, 4)
	SEQUENCE_FIELDS = [
		(LDAPString, 'objectName', '', False),
		(PartialAttributeList, 'attributes', lambda: [], False),
	]

	objectName: str
	attributes: typing.List[PartialAttribute]

class SearchResultDone(LDAPResult, ProtocolOp):
	BER_TAG = (1, True, 5)

class ModifyOperation(enum.Enum):
	# pylint: disable=invalid-name
	add = 0
	delete = 1
	replace = 2

class ModifyChange(asn1.Sequence):
	SEQUENCE_FIELDS = [
		(asn1.wrapenum(ModifyOperation), 'operation', None, False),
		(PartialAttribute, 'modification', None, False),
	]
	operation: ModifyOperation
	modification: PartialAttribute

class ModifyChanges(asn1.SequenceOf):
	SET_TYPE = ModifyChange

class ModifyRequest(asn1.Sequence, ProtocolOp):
	BER_TAG = (1, True, 6)
	SEQUENCE_FIELDS = [
		(LDAPString, 'object', None, False),
		(ModifyChanges, 'changes', None, False),
	]
	object: str
	changes: typing.List[ModifyChange]

class ModifyResponse(LDAPResult, ProtocolOp):
	BER_TAG = (1, True, 7)

class AddRequest(asn1.Sequence, ProtocolOp):
	BER_TAG = (1, True, 8)
	SEQUENCE_FIELDS = [
		(LDAPString, 'entry', None, False),
		(AttributeList, 'attributes', None, False),
	]
	entry: str
	attributes: Attribute

class AddResponse(LDAPResult, ProtocolOp):
	BER_TAG = (1, True, 9)

class DelRequest(asn1.Wrapper, ProtocolOp):
	BER_TAG = (1, False, 10)
	WRAPPED_ATTRIBUTE = 'dn'
	WRAPPED_TYPE = LDAPString
	WRAPPED_DEFAULT = None

	dn: str

class DelResponse(LDAPResult, ProtocolOp):
	BER_TAG = (1, True, 11)

class ModifyDNRequest(asn1.Sequence, ProtocolOp):
	BER_TAG = (1, True, 12)
	SEQUENCE_FIELDS = [
		(LDAPString, 'entry', None, False),
		(LDAPString, 'newrdn', None, False),
		(asn1.Boolean, 'deleteoldrdn', None, False),
		(asn1.retag(LDAPString, (2, False, 0)), 'newSuperior', None, True),
	]

	entry: str
	newrdn: str
	deleteoldrdn: bool
	newSuperior: str

class ModifyDNResponse(LDAPResult, ProtocolOp):
	BER_TAG = (1, True, 13)

class CompareRequest(asn1.Sequence, ProtocolOp):
	BER_TAG = (1, True, 14)
	SEQUENCE_FIELDS = [
		(LDAPString, 'entry', None, False),
		(AttributeValueAssertion, 'ava', None, False),
	]

	entry: str
	ava: AttributeValueAssertion

class CompareResponse(LDAPResult, ProtocolOp):
	BER_TAG = (1, True, 15)

class AbandonRequest(asn1.Wrapper, ProtocolOp):
	BER_TAG = (1, False, 16)
	WRAPPED_ATTRIBUTE = 'messageID'
	WRAPPED_TYPE = asn1.Integer
	WRAPPED_DEFAULT = None

	messageID: int

class ExtendedRequest(asn1.Sequence, ProtocolOp):
	BER_TAG = (1, True, 23)
	SEQUENCE_FIELDS = [
		(asn1.retag(LDAPOID, (2, False, 0)), 'requestName', None, True),
		(asn1.retag(asn1.OctetString, (2, False, 1)), 'requestValue', None, True),
	]

	requestName: str
	requestValue: bytes

class ExtendedResponse(asn1.Sequence, ProtocolOp):
	BER_TAG = (1, True, 24)
	SEQUENCE_FIELDS = [
		(asn1.wrapenum(LDAPResultCode), 'resultCode', None, False),
		(LDAPString, 'matchedDN', '', False),
		(LDAPString, 'diagnosticMessage', '', False),
		(asn1.retag(LDAPOID, (2, False, 10)), 'responseName', None, True),
		(asn1.retag(asn1.OctetString, (2, False, 11)), 'responseValue', None, True),
	]

	resultCode: LDAPResultCode
	matchedDN: str
	diagnosticMessage: str
	responseName: str
	responseValue: bytes

class IntermediateResponse(asn1.Sequence, ProtocolOp):
	BER_TAG = (1, True, 25)
	SEQUENCE_FIELDS = [
		(asn1.retag(LDAPOID, (2, False, 0)), 'responseName', None, True),
		(asn1.retag(asn1.OctetString, (2, False, 1)), 'responseValue', None, True),
	]

	responseName: str
	responseValue: bytes

class Control(asn1.Sequence):
	SEQUENCE_FIELDS = [
		(LDAPOID, 'controlType', None, False),
		(asn1.Boolean, 'criticality', None, True),
		(asn1.OctetString, 'controlValue', None, True),
	]

	controlType: str
	criticality: bool
	controlValue: bytes

class Controls(asn1.SequenceOf):
	BER_TAG = (2, True, 0)
	SET_TYPE = Control

class LDAPMessage(asn1.Sequence):
	SEQUENCE_FIELDS = [
		(asn1.Integer, 'messageID', None, False),
		(ProtocolOp, 'protocolOp', None, False),
		(Controls, 'controls', None, True)
	]

	messageID: int
	protocolOp: ProtocolOp
	controls: typing.List[Control]

class ShallowLDAPMessage(asn1.BERType):
	BER_TAG = (0, True, 16)
	# pylint: disable=invalid-name

	messageID: int
	protocolOpType: typing.Type[ProtocolOp]
	data: bytes

	def __init__(self, messageID=None, protocolOpType=None, data=None):
		self.messageID = messageID
		self.protocolOpType = protocolOpType
		self.data = data

	def decode(self):
		return LDAPMessage.from_ber(self.data)

	@classmethod
	def from_ber(cls, data):
		seq, rest = asn1.decode_ber(data)
		data = data[:len(data)-len(rest)]
		if seq.tag != cls.BER_TAG:
			raise ValueError()
		content = seq.content
		messageID, content = asn1.Integer.from_ber(content)
		op, content = asn1.decode_ber(content)
		for subcls in ProtocolOp.__subclasses__():
			if subcls.BER_TAG == op.tag:
				return cls(messageID, subcls, data), rest
		return cls(messageID, None, data), rest

	@classmethod
	def to_ber(cls, obj):
		if not isinstance(obj, cls):
			raise TypeError()
		return obj.data

# StartTLS Extended Operation (RFC4511)
STARTTLS_OID = '1.3.6.1.4.1.1466.20037'

# "Who am I?" Extended Operation (RFC4532)
WHOAMI_OID = '1.3.6.1.4.1.4203.1.11.3'

# Password Modify Extended Operation (RFC3062)
PASSWORD_MODIFY_OID = '1.3.6.1.4.1.4203.1.11.1'

class PasswdModifyRequestValue(asn1.Sequence):
	SEQUENCE_FIELDS = [
		(asn1.retag(LDAPString, (2, False, 0)), 'userIdentity', None, True),
		(asn1.retag(asn1.OctetString, (2, False, 1)), 'oldPasswd', None, True),
		(asn1.retag(asn1.OctetString, (2, False, 2)), 'newPasswd', None, True),
	]

	userIdentity: str
	oldPasswd: bytes
	newPasswd: bytes

class PasswdModifyResponseValue(asn1.Sequence):
	SEQUENCE_FIELDS = [
		(asn1.retag(asn1.OctetString, (2, False, 0)), 'genPasswd', None, True),
	]

	genPasswd: bytes

# LDAP Control Extension for Simple Paged Results Manipulation (RFC2696)
PAGED_RESULTS_OID = '1.2.840.113556.1.4.319'

class PagedResultsValue(asn1.Sequence):
	SEQUENCE_FIELDS = [
		(asn1.Integer, 'size', 0, False),
		(asn1.OctetString, 'cookie', b'', False)
	]

	size: int
	cookie: bytes



from .types import Schema
from .definitions import *
from . import syntaxes, matching_rules

RFC4512_ATTRIBUTE_TYPES = [
	"( 2.5.4.1 NAME 'aliasedObjectName' EQUALITY distinguishedNameMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.12 SINGLE-VALUE )",
	"( 2.5.4.0 NAME 'objectClass' EQUALITY objectIdentifierMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.38 )",
	"( 2.5.18.3 NAME 'creatorsName' EQUALITY distinguishedNameMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.12 SINGLE-VALUE NO-USER-MODIFICATION USAGE directoryOperation )",
	"( 2.5.18.1 NAME 'createTimestamp' EQUALITY generalizedTimeMatch ORDERING generalizedTimeOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.24 SINGLE-VALUE NO-USER-MODIFICATION USAGE directoryOperation )",
	"( 2.5.18.4 NAME 'modifiersName' EQUALITY distinguishedNameMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.12 SINGLE-VALUE NO-USER-MODIFICATION USAGE directoryOperation )",
	"( 2.5.18.2 NAME 'modifyTimestamp' EQUALITY generalizedTimeMatch ORDERING generalizedTimeOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.24 SINGLE-VALUE NO-USER-MODIFICATION USAGE directoryOperation )",
	"( 2.5.21.9 NAME 'structuralObjectClass' EQUALITY objectIdentifierMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.38 SINGLE-VALUE NO-USER-MODIFICATION USAGE directoryOperation ) ",
	"( 2.5.21.10 NAME 'governingStructureRule' EQUALITY integerMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE NO-USER-MODIFICATION USAGE directoryOperation )",
	"( 2.5.18.10 NAME 'subschemaSubentry' EQUALITY distinguishedNameMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.12 SINGLE-VALUE NO-USER-MODIFICATION USAGE directoryOperation )",
	"( 2.5.21.6 NAME 'objectClasses' EQUALITY objectIdentifierFirstComponentMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.37 USAGE directoryOperation )",
	"( 2.5.21.5 NAME 'attributeTypes' EQUALITY objectIdentifierFirstComponentMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.3 USAGE directoryOperation )",
	"( 2.5.21.4 NAME 'matchingRules' EQUALITY objectIdentifierFirstComponentMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.30 USAGE directoryOperation )",
	"( 2.5.21.8 NAME 'matchingRuleUse' EQUALITY objectIdentifierFirstComponentMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.31 USAGE directoryOperation )",
	"( 1.3.6.1.4.1.1466.101.120.16 NAME 'ldapSyntaxes' EQUALITY objectIdentifierFirstComponentMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.54 USAGE directoryOperation )",
	"( 2.5.21.2 NAME 'dITContentRules' EQUALITY objectIdentifierFirstComponentMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.16 USAGE directoryOperation )",
	"( 2.5.21.1 NAME 'dITStructureRules' EQUALITY integerFirstComponentMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.17 USAGE directoryOperation )",
	"( 2.5.21.7 NAME 'nameForms' EQUALITY objectIdentifierFirstComponentMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.35 USAGE directoryOperation )",
	"( 1.3.6.1.4.1.1466.101.120.6 NAME 'altServer' SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 USAGE dSAOperation )",
	"( 1.3.6.1.4.1.1466.101.120.5 NAME 'namingContexts' SYNTAX 1.3.6.1.4.1.1466.115.121.1.12 USAGE dSAOperation )",
	"( 1.3.6.1.4.1.1466.101.120.13 NAME 'supportedControl' SYNTAX 1.3.6.1.4.1.1466.115.121.1.38 USAGE dSAOperation )",
	"( 1.3.6.1.4.1.1466.101.120.7 NAME 'supportedExtension' SYNTAX 1.3.6.1.4.1.1466.115.121.1.38 USAGE dSAOperation )",
	"( 1.3.6.1.4.1.4203.1.3.5 NAME 'supportedFeatures' EQUALITY objectIdentifierMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.38 USAGE dSAOperation )",
	"( 1.3.6.1.4.1.1466.101.120.15 NAME 'supportedLDAPVersion' SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 USAGE dSAOperation )",
	"( 1.3.6.1.4.1.1466.101.120.14 NAME 'supportedSASLMechanisms' SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 USAGE dSAOperation )",
]
RFC4512_OBJECT_CLASSES = [
	"( 2.5.6.0 NAME 'top' ABSTRACT MUST objectClass )",
	"( 2.5.6.1 NAME 'alias' SUP top STRUCTURAL MUST aliasedObjectName )",
	"( 1.3.6.1.4.1.1466.101.120.111 NAME 'extensibleObject' SUP top AUXILIARY )",
	"( 2.5.20.1 NAME 'subschema' AUXILIARY MAY ( dITStructureRules $ nameForms $ ditContentRules $ objectClasses $ attributeTypes $ matchingRules $ matchingRuleUse ) )",
]
CORE_SCHEMA = RFC4512_SCHEMA = Schema(syntax_definitions=syntaxes.ALL, matching_rule_definitions=matching_rules.ALL, attribute_type_definitions=RFC4512_ATTRIBUTE_TYPES, object_class_definitions=RFC4512_OBJECT_CLASSES)

RFC4519_ATTRIBUTE_TYPES = [
	"( 2.5.4.15 NAME 'businessCategory' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.5.4.6 NAME ( 'c' 'countryName' ) SUP name SYNTAX 1.3.6.1.4.1.1466.115.121.1.11 SINGLE-VALUE )",
	"( 2.5.4.3 NAME ( 'cn' 'commonName' ) SUP name )",
	"( 0.9.2342.19200300.100.1.25 NAME ( 'dc' 'domainComponent' ) EQUALITY caseIgnoreIA5Match SUBSTR caseIgnoreIA5SubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 SINGLE-VALUE )",
	"( 2.5.4.13 NAME 'description' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.5.4.27 NAME 'destinationIndicator' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.44 )",
	"( 2.5.4.49 NAME 'distinguishedName' EQUALITY distinguishedNameMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.12 )",
	"( 2.5.4.46 NAME 'dnQualifier' EQUALITY caseIgnoreMatch ORDERING caseIgnoreOrderingMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.44 )",
	"( 2.5.4.47 NAME 'enhancedSearchGuide' SYNTAX 1.3.6.1.4.1.1466.115.121.1.21 )",
	"( 2.5.4.23 NAME 'facsimileTelephoneNumber' SYNTAX 1.3.6.1.4.1.1466.115.121.1.22 )",
	"( 2.5.4.44 NAME 'generationQualifier' SUP name )",
	"( 2.5.4.42 NAME ( 'givenName' 'gn' ) SUP name )",
	"( 2.5.4.51 NAME 'houseIdentifier' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.5.4.43 NAME 'initials' SUP name )",
	"( 2.5.4.25 NAME 'internationalISDNNumber' EQUALITY numericStringMatch SUBSTR numericStringSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.36 )",
	"( 2.5.4.7 NAME ( 'l' 'localityName' ) SUP name )",
	"( 2.5.4.31 NAME 'member' SUP distinguishedName )",
	"( 2.5.4.41 NAME 'name' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.5.4.10 NAME ( 'o' 'organizationName' ) SUP name )",
	"( 2.5.4.11 NAME ( 'ou' 'organizationalUnitName' ) SUP name )",
	"( 2.5.4.32 NAME 'owner' SUP distinguishedName )",
	"( 2.5.4.19 NAME 'physicalDeliveryOfficeName' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.5.4.16 NAME 'postalAddress' EQUALITY caseIgnoreListMatch SUBSTR caseIgnoreListSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.41 )",
	"( 2.5.4.17 NAME 'postalCode' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.5.4.18 NAME 'postOfficeBox' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.5.4.28 NAME 'preferredDeliveryMethod' SYNTAX 1.3.6.1.4.1.1466.115.121.1.14 SINGLE-VALUE )",
	"( 2.5.4.26 NAME 'registeredAddress' SUP postalAddress SYNTAX 1.3.6.1.4.1.1466.115.121.1.41 )",
	"( 2.5.4.33 NAME 'roleOccupant' SUP distinguishedName )",
	"( 2.5.4.14 NAME 'searchGuide' SYNTAX 1.3.6.1.4.1.1466.115.121.1.25 )",
	"( 2.5.4.34 NAME 'seeAlso' SUP distinguishedName )",
	"( 2.5.4.5 NAME 'serialNumber' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.44 )",
	"( 2.5.4.4 NAME ( 'sn' 'surname' ) SUP name )",
	"( 2.5.4.8 NAME 'st' SUP name )",
	"( 2.5.4.9 NAME 'street' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.5.4.20 NAME 'telephoneNumber' EQUALITY telephoneNumberMatch SUBSTR telephoneNumberSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.50 )",
	"( 2.5.4.22 NAME 'teletexTerminalIdentifier' SYNTAX 1.3.6.1.4.1.1466.115.121.1.51 )",
	"( 2.5.4.21 NAME 'telexNumber' SYNTAX 1.3.6.1.4.1.1466.115.121.1.52 )",
	"( 2.5.4.12 NAME 'title' SUP name )",
	"( 0.9.2342.19200300.100.1.1 NAME ( 'uid' 'userid' ) EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.5.4.50 NAME 'uniqueMember' EQUALITY uniqueMemberMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.34 )",
	"( 2.5.4.35 NAME 'userPassword' EQUALITY octetStringMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.40 )",
	"( 2.5.4.24 NAME 'x121Address' EQUALITY numericStringMatch SUBSTR numericStringSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.36 )",
	"( 2.5.4.45 NAME 'x500UniqueIdentifier' EQUALITY bitStringMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.6 )",
]
RFC4519_OBJECT_CLASSES = [
	"( 2.5.6.11 NAME 'applicationProcess' SUP top STRUCTURAL MUST cn MAY ( seeAlso $ ou $ l $ description ) )",
	"( 2.5.6.2 NAME 'country' SUP top STRUCTURAL MUST c MAY ( searchGuide $ description ) )",
	"( 1.3.6.1.4.1.1466.344 NAME 'dcObject' SUP top AUXILIARY MUST dc )",
	"( 2.5.6.14 NAME 'device' SUP top STRUCTURAL MUST cn MAY ( serialNumber $ seeAlso $ owner $ ou $ o $ l $ description ) )",
	"( 2.5.6.9 NAME 'groupOfNames' SUP top STRUCTURAL MUST ( member $ cn ) MAY ( businessCategory $ seeAlso $ owner $ ou $ o $ description ) )",
	"( 2.5.6.17 NAME 'groupOfUniqueNames' SUP top STRUCTURAL MUST ( uniqueMember $ cn ) MAY ( businessCategory $ seeAlso $ owner $ ou $ o $ description ) )",
	"( 2.5.6.3 NAME 'locality' SUP top STRUCTURAL MAY ( street $ seeAlso $ searchGuide $ st $ l $ description ) )",
	"( 2.5.6.4 NAME 'organization' SUP top STRUCTURAL MUST o MAY ( userPassword $ searchGuide $ seeAlso $ businessCategory $ x121Address $ registeredAddress $ destinationIndicator $ preferredDeliveryMethod $ telexNumber $ teletexTerminalIdentifier $ telephoneNumber $ internationalISDNNumber $ facsimileTelephoneNumber $ street $ postOfficeBox $ postalCode $ postalAddress $ physicalDeliveryOfficeName $ st $ l $ description ) )",
	"( 2.5.6.7 NAME 'organizationalPerson' SUP person STRUCTURAL MAY ( title $ x121Address $ registeredAddress $ destinationIndicator $ preferredDeliveryMethod $ telexNumber $ teletexTerminalIdentifier $ telephoneNumber $ internationalISDNNumber $ facsimileTelephoneNumber $ street $ postOfficeBox $ postalCode $ postalAddress $ physicalDeliveryOfficeName $ ou $ st $ l ) )",
	"( 2.5.6.8 NAME 'organizationalRole' SUP top STRUCTURAL MUST cn MAY ( x121Address $ registeredAddress $ destinationIndicator $ preferredDeliveryMethod $ telexNumber $ teletexTerminalIdentifier $ telephoneNumber $ internationalISDNNumber $ facsimileTelephoneNumber $ seeAlso $ roleOccupant $ preferredDeliveryMethod $ street $ postOfficeBox $ postalCode $ postalAddress $ physicalDeliveryOfficeName $ ou $ st $ l $ description ) )",
	"( 2.5.6.5 NAME 'organizationalUnit' SUP top STRUCTURAL MUST ou MAY ( businessCategory $ description $ destinationIndicator $ facsimileTelephoneNumber $ internationalISDNNumber $ l $ physicalDeliveryOfficeName $ postalAddress $ postalCode $ postOfficeBox $ preferredDeliveryMethod $ registeredAddress $ searchGuide $ seeAlso $ st $ street $ telephoneNumber $ teletexTerminalIdentifier $ telexNumber $ userPassword $ x121Address ) )",
	"( 2.5.6.6 NAME 'person' SUP top STRUCTURAL MUST ( sn $ cn ) MAY ( userPassword $ telephoneNumber $ seeAlso $ description ) )",
	"( 2.5.6.10 NAME 'residentialPerson' SUP person STRUCTURAL MUST l MAY ( businessCategory $ x121Address $ registeredAddress $ destinationIndicator $ preferredDeliveryMethod $ telexNumber $ teletexTerminalIdentifier $ telephoneNumber $ internationalISDNNumber $ facsimileTelephoneNumber $ preferredDeliveryMethod $ street $ postOfficeBox $ postalCode $ postalAddress $ physicalDeliveryOfficeName $ st $ l ) )",
	"( 1.3.6.1.1.3.1 NAME 'uidObject' SUP top AUXILIARY MUST uid )",
]
RFC4519_SCHEMA = CORE_SCHEMA.extend(attribute_type_definitions=RFC4519_ATTRIBUTE_TYPES, object_class_definitions=RFC4519_OBJECT_CLASSES)

RFC4523_ATTRIBUTE_TYPES = [
	"( 2.5.4.36 NAME 'userCertificate' DESC 'X.509 user certificate' EQUALITY certificateExactMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.8 )",
	"( 2.5.4.37 NAME 'cACertificate' DESC 'X.509 CA certificate' EQUALITY certificateExactMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.8 )",
	"( 2.5.4.40 NAME 'crossCertificatePair' DESC 'X.509 cross certificate pair' EQUALITY certificatePairExactMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.10 )",
	"( 2.5.4.39 NAME 'certificateRevocationList' DESC 'X.509 certificate revocation list' EQUALITY certificateListExactMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.9 )",
	"( 2.5.4.38 NAME 'authorityRevocationList' DESC 'X.509 authority revocation list' EQUALITY certificateListExactMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.9 )",
	"( 2.5.4.53 NAME 'deltaRevocationList' DESC 'X.509 delta revocation list' EQUALITY certificateListExactMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.9 )",
	"( 2.5.4.52 NAME 'supportedAlgorithms' DESC 'X.509 supported algorithms' EQUALITY algorithmIdentifierMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.49 )",
]
RFC4523_OBJECT_CLASSES = [
	"( 2.5.6.21 NAME 'pkiUser' DESC 'X.509 PKI User' SUP top AUXILIARY MAY userCertificate )",
	"( 2.5.6.22 NAME 'pkiCA' DESC 'X.509 PKI Certificate Authority' SUP top AUXILIARY MAY ( cACertificate $ certificateRevocationList $ authorityRevocationList $ crossCertificatePair ) )",
	"( 2.5.6.19 NAME 'cRLDistributionPoint' DESC 'X.509 CRL distribution point' SUP top STRUCTURAL MUST cn MAY ( certificateRevocationList $ authorityRevocationList $ deltaRevocationList ) )",
	"( 2.5.6.23 NAME 'deltaCRL' DESC 'X.509 delta CRL' SUP top AUXILIARY MAY deltaRevocationList )",
	"( 2.5.6.15 NAME 'strongAuthenticationUser' DESC 'X.521 strong authentication user' SUP top AUXILIARY MUST userCertificate )",
	"( 2.5.6.18 NAME 'userSecurityInformation' DESC 'X.521 user security information' SUP top AUXILIARY MAY ( supportedAlgorithms ) )",
	"( 2.5.6.16 NAME 'certificationAuthority' DESC 'X.509 certificate authority' SUP top AUXILIARY MUST ( authorityRevocationList $ certificateRevocationList $ cACertificate ) MAY crossCertificatePair )",
	"( 2.5.6.16.2 NAME 'certificationAuthority-V2' DESC 'X.509 certificate authority, version 2' SUP certificationAuthority AUXILIARY MAY deltaRevocationList )",
]
RFC4523_SCHEMA = RFC4519_SCHEMA.extend(attribute_type_definitions=RFC4523_ATTRIBUTE_TYPES, object_class_definitions=RFC4523_OBJECT_CLASSES)

RFC4524_ATTRIBUTE_TYPES = [
	"( 0.9.2342.19200300.100.1.37 NAME 'associatedDomain' EQUALITY caseIgnoreIA5Match SUBSTR caseIgnoreIA5SubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 )",
	"( 0.9.2342.19200300.100.1.38 NAME 'associatedName' EQUALITY distinguishedNameMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.12 )",
	"( 0.9.2342.19200300.100.1.48 NAME 'buildingName' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{256} )",
	"( 0.9.2342.19200300.100.1.43 NAME 'co' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 0.9.2342.19200300.100.1.14 NAME 'documentAuthor' EQUALITY distinguishedNameMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.12 )",
	"( 0.9.2342.19200300.100.1.11 NAME 'documentIdentifier' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{256} )",
	"( 0.9.2342.19200300.100.1.15 NAME 'documentLocation' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{256} )",
	"( 0.9.2342.19200300.100.1.56 NAME 'documentPublisher' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 0.9.2342.19200300.100.1.12 NAME 'documentTitle' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{256} )",
	"( 0.9.2342.19200300.100.1.13 NAME 'documentVersion' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{256} )",
	"( 0.9.2342.19200300.100.1.5 NAME 'drink' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{256} )",
	"( 0.9.2342.19200300.100.1.20 NAME 'homePhone' EQUALITY telephoneNumberMatch SUBSTR telephoneNumberSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.50 )",
	"( 0.9.2342.19200300.100.1.39 NAME 'homePostalAddress' EQUALITY caseIgnoreListMatch SUBSTR caseIgnoreListSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.41 )",
	"( 0.9.2342.19200300.100.1.9 NAME 'host' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{256} )",
	"( 0.9.2342.19200300.100.1.4 NAME 'info' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{2048} )",
	"( 0.9.2342.19200300.100.1.3 NAME 'mail' EQUALITY caseIgnoreIA5Match SUBSTR caseIgnoreIA5SubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.26{256} )",
	"( 0.9.2342.19200300.100.1.10 NAME 'manager' EQUALITY distinguishedNameMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.12 )",
	"( 0.9.2342.19200300.100.1.41 NAME 'mobile' EQUALITY telephoneNumberMatch SUBSTR telephoneNumberSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.50 )",
	"( 0.9.2342.19200300.100.1.45 NAME 'organizationalStatus' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{256} )",
	"( 0.9.2342.19200300.100.1.42 NAME 'pager' EQUALITY telephoneNumberMatch SUBSTR telephoneNumberSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.50 )",
	"( 0.9.2342.19200300.100.1.40 NAME 'personalTitle' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{256} )",
	"( 0.9.2342.19200300.100.1.6 NAME 'roomNumber' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{256} )",
	"( 0.9.2342.19200300.100.1.21 NAME 'secretary' EQUALITY distinguishedNameMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.12 )",
	"( 0.9.2342.19200300.100.1.44 NAME 'uniqueIdentifier' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{256} )",
	"( 0.9.2342.19200300.100.1.8 NAME 'userClass' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{256} )",
]
RFC4524_OBJECT_CLASSES = [
	"( 0.9.2342.19200300.100.4.5 NAME 'account' SUP top STRUCTURAL MUST uid MAY ( description $ seeAlso $ l $ o $ ou $ host ) )",
	"( 0.9.2342.19200300.100.4.6 NAME 'document' SUP top STRUCTURAL MUST documentIdentifier MAY ( cn $ description $ seeAlso $ l $ o $ ou $ documentTitle $ documentVersion $ documentAuthor $ documentLocation $ documentPublisher ) )",
	"( 0.9.2342.19200300.100.4.9 NAME 'documentSeries' SUP top STRUCTURAL MUST cn MAY ( description $ l $ o $ ou $ seeAlso $ telephonenumber ) )",
	"( 0.9.2342.19200300.100.4.13 NAME 'domain' SUP top STRUCTURAL MUST dc MAY ( userPassword $ searchGuide $ seeAlso $ businessCategory $ x121Address $ registeredAddress $ destinationIndicator $ preferredDeliveryMethod $ telexNumber $ teletexTerminalIdentifier $ telephoneNumber $ internationaliSDNNumber $ facsimileTelephoneNumber $ street $ postOfficeBox $ postalCode $ postalAddress $ physicalDeliveryOfficeName $ st $ l $ description $ o $ associatedName ) )",
	"( 0.9.2342.19200300.100.4.17 NAME 'domainRelatedObject' SUP top AUXILIARY MUST associatedDomain )",
	"( 0.9.2342.19200300.100.4.18 NAME 'friendlyCountry' SUP country STRUCTURAL MUST co )",
	"( 0.9.2342.19200300.100.4.14 NAME 'rFC822localPart' SUP domain STRUCTURAL MAY ( cn $ description $ destinationIndicator $ facsimileTelephoneNumber $ internationaliSDNNumber $ physicalDeliveryOfficeName $ postalAddress $ postalCode $ postOfficeBox $ preferredDeliveryMethod $ registeredAddress $ seeAlso $ sn $ street $ telephoneNumber $ teletexTerminalIdentifier $ telexNumber $ x121Address ) )",
	"( 0.9.2342.19200300.100.4.7 NAME 'room' SUP top STRUCTURAL MUST cn MAY ( roomNumber $ description $ seeAlso $ telephoneNumber ) )",
	"( 0.9.2342.19200300.100.4.19 NAME 'simpleSecurityObject' SUP top AUXILIARY MUST userPassword )",
]
COSINE_SCHEMA = RFC4524_SCHEMA = RFC4519_SCHEMA.extend(attribute_type_definitions=RFC4524_ATTRIBUTE_TYPES, object_class_definitions=RFC4524_OBJECT_CLASSES)

RFC3112_ATTRIBUTE_TYPES = [
	"( 1.3.6.1.4.1.4203.1.3.3 NAME 'supportedAuthPasswordSchemes' DESC 'supported password storage schemes' EQUALITY caseExactIA5Match SYNTAX 1.3.6.1.4.1.1466.115.121.1.26{32} USAGE dSAOperation )",
	"( 1.3.6.1.4.1.4203.1.3.4 NAME 'authPassword' DESC 'password authentication information' EQUALITY 1.3.6.1.4.1.4203.1.2.2 SYNTAX 1.3.6.1.4.1.4203.1.1.2 )",
]
RFC3112_OBJECT_CLASSES = [
	"( 1.3.6.1.4.1.4203.1.4.7 NAME 'authPasswordObject' DESC 'authentication password mix in class' AUXILIARY MAY authPassword )",
]
RFC3112_SCHEMA = CORE_SCHEMA.extend(attribute_type_definitions=RFC3112_ATTRIBUTE_TYPES, object_class_definitions=RFC3112_OBJECT_CLASSES)

RFC2079_ATTRIBUTE_TYPES = [
	"( 1.3.6.1.4.1.250.1.57 NAME 'labeledURI' DESC 'Uniform Resource Identifier with optional label' EQUALITY caseExactMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
]
RFC2079_OBJECT_CLASSES = [
	"( 1.3.6.1.4.1.250.3.15 NAME 'labeledURIObject' DESC 'object that contains the URI attribute type' SUP top AUXILIARY MAY labeledURI )",
]
RFC2079_SCHEMA = CORE_SCHEMA.extend(attribute_type_definitions=RFC2079_ATTRIBUTE_TYPES, object_class_definitions=RFC2079_OBJECT_CLASSES)

RFC2798_ATTRIBUTE_TYPES = [
	# Originally from RFC1274, but updated and used by RFC2798
	"( 0.9.2342.19200300.100.1.55 NAME 'audio' EQUALITY octetStringMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.40{250000} )",
	# Also from RFC1274 and updated by RFC2798, but lacks SYNTAX, so "Fax" is added here
	"( 0.9.2342.19200300.100.1.7 NAME 'photo' SYNTAX 1.3.6.1.4.1.1466.115.121.1.23 )",

	"( 2.16.840.1.113730.3.1.1 NAME 'carLicense' DESC 'vehicle license or registration plate' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.16.840.1.113730.3.1.2 NAME 'departmentNumber' DESC 'identifies a department within an organization' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 2.16.840.1.113730.3.1.241 NAME 'displayName' DESC 'preferred name of a person to be used when displaying entries' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 SINGLE-VALUE )",
	"( 2.16.840.1.113730.3.1.3 NAME 'employeeNumber' DESC 'numerically identifies an employee within an organization' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 SINGLE-VALUE )",
	"( 2.16.840.1.113730.3.1.4 NAME 'employeeType' DESC 'type of employment for a person' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 0.9.2342.19200300.100.1.60 NAME 'jpegPhoto' DESC 'a JPEG image' SYNTAX 1.3.6.1.4.1.1466.115.121.1.28 )",
	"( 2.16.840.1.113730.3.1.39 NAME 'preferredLanguage' DESC 'preferred written or spoken language for a person' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 SINGLE-VALUE )",
	"( 2.16.840.1.113730.3.1.40 NAME 'userSMIMECertificate' DESC 'signed message used to support S/MIME' SYNTAX 1.3.6.1.4.1.1466.115.121.1.5 )",
	"( 2.16.840.1.113730.3.1.216 NAME 'userPKCS12' DESC 'PKCS #12 PFX PDU for exchange of personal identity information' SYNTAX 1.3.6.1.4.1.1466.115.121.1.5 )",
]
RFC2798_OBJECT_CLASSES = [
	"( 2.16.840.1.113730.3.2.2 NAME 'inetOrgPerson' SUP organizationalPerson STRUCTURAL MAY ( audio $ businessCategory $ carLicense $ departmentNumber $ displayName $ employeeNumber $ employeeType $ givenName $ homePhone $ homePostalAddress $ initials $ jpegPhoto $ labeledURI $ mail $ manager $ mobile $ o $ pager $ photo $ roomNumber $ secretary $ uid $ userCertificate $ x500uniqueIdentifier $ preferredLanguage $ userSMIMECertificate $ userPKCS12 ) )",
]
INETORG_SCHMEA = RFC2798_SCHEMA = (RFC4524_SCHEMA|RFC2079_SCHEMA|RFC4523_SCHEMA).extend(attribute_type_definitions=RFC2798_ATTRIBUTE_TYPES, object_class_definitions=RFC2798_OBJECT_CLASSES)

RFC2307BIS_ATTRIBUTE_TYPES = [
	"( 1.3.6.1.1.1.1.0 NAME 'uidNumber' DESC 'An integer uniquely identifying a user in an administrative domain' EQUALITY integerMatch ORDERING integerOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.1 NAME 'gidNumber' DESC 'An integer uniquely identifying a group in an administrative domain' EQUALITY integerMatch ORDERING integerOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.2 NAME 'gecos' DESC 'The GECOS field; the common name' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.3 NAME 'homeDirectory' DESC 'The absolute path to the home directory' EQUALITY caseExactIA5Match SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.4 NAME 'loginShell' DESC 'The path to the login shell' EQUALITY caseExactIA5Match SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.5 NAME 'shadowLastChange' EQUALITY integerMatch ORDERING integerOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.6 NAME 'shadowMin' EQUALITY integerMatch ORDERING integerOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.7 NAME 'shadowMax' EQUALITY integerMatch ORDERING integerOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.8 NAME 'shadowWarning' EQUALITY integerMatch ORDERING integerOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.9 NAME 'shadowInactive' EQUALITY integerMatch ORDERING integerOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.10 NAME 'shadowExpire' EQUALITY integerMatch ORDERING integerOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.11 NAME 'shadowFlag' EQUALITY integerMatch ORDERING integerOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.12 NAME 'memberUid' EQUALITY caseExactMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 1.3.6.1.1.1.1.13 NAME 'memberNisNetgroup' EQUALITY caseExactMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 1.3.6.1.1.1.1.14 NAME 'nisNetgroupTriple' DESC 'Netgroup triple' EQUALITY caseIgnoreMatch SUBSTR caseIgnoreSubstringsMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 1.3.6.1.1.1.1.15 NAME 'ipServicePort' DESC 'Service port number' EQUALITY integerMatch ORDERING integerOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.16 NAME 'ipServiceProtocol' DESC 'Service protocol name' EQUALITY caseIgnoreMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 )",
	"( 1.3.6.1.1.1.1.17 NAME 'ipProtocolNumber' DESC 'IP protocol number' EQUALITY integerMatch ORDERING integerOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.18 NAME 'oncRpcNumber' DESC 'ONC RPC number' EQUALITY integerMatch ORDERING integerOrderingMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.27 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.19 NAME 'ipHostNumber' DESC 'IPv4 addresses as a dotted decimal omitting leading zeros or IPv6 addresses as defined in RFC2373' EQUALITY caseIgnoreIA5Match SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 )",
	"( 1.3.6.1.1.1.1.20 NAME 'ipNetworkNumber' DESC 'IP network omitting leading zeros, eg. 192.168' EQUALITY caseIgnoreIA5Match SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.21 NAME 'ipNetmaskNumber' DESC 'IP netmask omitting leading zeros, eg. 255.255.255.0' EQUALITY caseIgnoreIA5Match SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.22 NAME 'macAddress' DESC 'MAC address in maximal, colon separated hex notation, eg. 00:00:92:90:ee:e2' EQUALITY caseIgnoreIA5Match SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 )",
	"( 1.3.6.1.1.1.1.23 NAME 'bootParameter' DESC 'rpc.bootparamd parameter' EQUALITY caseExactIA5Match SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 )",
	"( 1.3.6.1.1.1.1.24 NAME 'bootFile' DESC 'Boot image name' EQUALITY caseExactIA5Match SYNTAX 1.3.6.1.4.1.1466.115.121.1.26 )",
	"( 1.3.6.1.1.1.1.26 NAME 'nisMapName' DESC 'Name of a generic NIS map' EQUALITY caseIgnoreMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{64} )",
	"( 1.3.6.1.1.1.1.27 NAME 'nisMapEntry' DESC 'A generic NIS entry' EQUALITY caseExactMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15{1024} SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.28 NAME 'nisPublicKey' DESC 'NIS public key' EQUALITY octetStringMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.40 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.29 NAME 'nisSecretKey' DESC 'NIS secret key' EQUALITY octetStringMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.40 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.30 NAME 'nisDomain' DESC 'NIS domain' EQUALITY caseIgnoreIA5Match SYNTAX 1.3.6.1.4.1.1466.115.121.1.26{256} )",
	"( 1.3.6.1.1.1.1.31 NAME 'automountMapName' DESC 'automount Map Name' EQUALITY caseExactMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.32 NAME 'automountKey' DESC 'Automount Key value' EQUALITY caseExactMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 SINGLE-VALUE )",
	"( 1.3.6.1.1.1.1.33 NAME 'automountInformation' DESC 'Automount information' EQUALITY caseExactMatch SYNTAX 1.3.6.1.4.1.1466.115.121.1.15 SINGLE-VALUE )",
]
RFC2307BIS_OBJECT_CLASSES = [
	"( 1.3.6.1.1.1.2.0 NAME 'posixAccount' DESC 'Abstraction of an account with POSIX attributes' SUP top AUXILIARY MUST ( cn $ uid $ uidNumber $ gidNumber $ homeDirectory ) MAY ( authPassword $ userPassword $ loginShell $ gecos $ description ) )",
	"( 1.3.6.1.1.1.2.1 NAME 'shadowAccount' DESC 'Additional attributes for shadow passwords' SUP top AUXILIARY MUST uid MAY ( authPassword $ userPassword $ description $ shadowLastChange $ shadowMin $ shadowMax $ shadowWarning $ shadowInactive $ shadowExpire $ shadowFlag ) )",
	"( 1.3.6.1.1.1.2.2 NAME 'posixGroup' DESC 'Abstraction of a group of accounts' SUP top AUXILIARY MUST gidNumber MAY ( authPassword $ userPassword $ memberUid $ description ) )",
	"( 1.3.6.1.1.1.2.3 NAME 'ipService' DESC 'Abstraction an Internet Protocol service.  Maps an IP port and protocol (such as tcp or udp) to one or more names; the distinguished value of the cn attribute denotes the service\\27s canonical name' SUP top STRUCTURAL MUST ( cn $ ipServicePort $ ipServiceProtocol ) MAY description )",
	"( 1.3.6.1.1.1.2.4 NAME 'ipProtocol' DESC 'Abstraction of an IP protocol. Maps a protocol number to one or more names. The distinguished value of the cn attribute denotes the protocol canonical name' SUP top STRUCTURAL MUST ( cn $ ipProtocolNumber ) MAY description )",
	"( 1.3.6.1.1.1.2.5 NAME 'oncRpc' DESC 'Abstraction of an Open Network Computing (ONC) [RFC1057] Remote Procedure Call (RPC) binding.  This class maps an ONC RPC number to a name.  The distinguished value of the cn attribute denotes the RPC service canonical name' SUP top STRUCTURAL MUST ( cn $ oncRpcNumber ) MAY description )",
	"( 1.3.6.1.1.1.2.6 NAME 'ipHost' DESC 'Abstraction of a host, an IP device. The distinguished value of the cn attribute denotes the host\\27s canonical name. Device SHOULD be used as a structural class' SUP top AUXILIARY MUST ( cn $ ipHostNumber ) MAY ( authPassword $ userPassword $ l $ description $ manager ) )",
	"( 1.3.6.1.1.1.2.7 NAME 'ipNetwork' DESC 'Abstraction of a network. The distinguished value of the cn attribute denotes the network canonical name' SUP top STRUCTURAL MUST ipNetworkNumber MAY ( cn $ ipNetmaskNumber $ l $ description $ manager ) )",
	"( 1.3.6.1.1.1.2.8 NAME 'nisNetgroup' DESC 'Abstraction of a netgroup. May refer to other netgroups' SUP top STRUCTURAL MUST cn MAY ( nisNetgroupTriple $ memberNisNetgroup $ description ) )",
	"( 1.3.6.1.1.1.2.9 NAME 'nisMap' DESC 'A generic abstraction of a NIS map' SUP top STRUCTURAL MUST nisMapName MAY description )",
	"( 1.3.6.1.1.1.2.10 NAME 'nisObject' DESC 'An entry in a NIS map' SUP top STRUCTURAL MUST ( cn $ nisMapEntry $ nisMapName ) )",
	"( 1.3.6.1.1.1.2.11 NAME 'ieee802Device' DESC 'A device with a MAC address; device SHOULD be used as a structural class' SUP top AUXILIARY MAY macAddress )",
	"( 1.3.6.1.1.1.2.12 NAME 'bootableDevice' DESC 'A device with boot parameters; device SHOULD be used as a structural class' SUP top AUXILIARY MAY ( bootFile $ bootParameter ) )",
	"( 1.3.6.1.1.1.2.14 NAME 'nisKeyObject' DESC 'An object with a public and secret key' SUP top AUXILIARY MUST ( cn $ nisPublicKey $ nisSecretKey ) MAY ( uidNumber $ description ) )",
	"( 1.3.6.1.1.1.2.15 NAME 'nisDomainObject' DESC 'Associates a NIS domain with a naming context' SUP top AUXILIARY MUST nisDomain )",
	"( 1.3.6.1.1.1.2.16 NAME 'automountMap' SUP top STRUCTURAL MUST ( automountMapName ) MAY description )",
	"( 1.3.6.1.1.1.2.17 NAME 'automount' DESC 'Automount information' SUP top STRUCTURAL MUST ( automountKey $ automountInformation ) MAY description )",
	"( 1.3.6.1.1.1.2.18 NAME 'groupOfMembers' DESC 'A group with members (DNs)' SUP top STRUCTURAL MUST cn MAY ( businessCategory $ seeAlso $ owner $ ou $ o $ description $ member ) )",
]
RFC2307BIS_SCHEMA = (RFC4524_SCHEMA|RFC3112_SCHEMA).extend(attribute_type_definitions=RFC2307BIS_ATTRIBUTE_TYPES, object_class_definitions=RFC2307BIS_OBJECT_CLASSES)



import re
import datetime

from .definitions import SyntaxDefinition
from .. import dn, exceptions

class BytesSyntaxDefinition(SyntaxDefinition):
	def encode(self, schema, value):
		return value

	def decode(self, schema, raw_value):
		return raw_value

class StringSyntaxDefinition(SyntaxDefinition):
	def __init__(self, oid, encoding='utf-8', re_pattern='.*', **kwargs):
		super().__init__(oid, **kwargs)
		self.encoding = encoding
		self.re_pattern = re.compile(re_pattern)

	def encode(self, schema, value):
		return value.encode(self.encoding)

	def decode(self, schema, raw_value):
		try:
			value = raw_value.decode(self.encoding)
		except UnicodeDecodeError as exc:
			raise exceptions.LDAPInvalidAttributeSyntax(exc.reason) from exc
		if not re.fullmatch(self.re_pattern, value):
			raise exceptions.LDAPInvalidAttributeSyntax()
		return value

class IntegerSyntaxDefinition(StringSyntaxDefinition):
	def __init__(self, oid, **kwargs):
		super().__init__(oid, encoding='ascii', re_pattern='([0-9]|-?[1-9][0-9]+)', **kwargs)

	def encode(self, schema, value):
		return super().encode(schema, str(value))

	def decode(self, schema, raw_value):
		try:
			return int(super().decode(schema, raw_value))
		except ValueError as exc:
			raise exceptions.LDAPInvalidAttributeSyntax() from exc

class SchemaElementSyntaxDefinition(SyntaxDefinition):
	def encode(self, schema, value):
		return str(value).encode('utf8')

class BooleanSyntaxDefinition(SyntaxDefinition):
	def encode(self, schema, value):
		return b'TRUE' if value else b'FALSE'

	def decode(self, schema, raw_value):
		if raw_value == b'TRUE':
			return True
		elif raw_value == b'FALSE':
			return False
		else:
			raise exceptions.LDAPInvalidAttributeSyntax()

class DNSyntaxDefinition(SyntaxDefinition):
	def encode(self, schema, value):
		return str(value).encode('utf8')

	def decode(self, schema, raw_value):
		try:
			return dn.DN.from_str(schema, raw_value.decode('utf8'))
		except (UnicodeDecodeError, TypeError, ValueError) as exc:
			raise exceptions.LDAPInvalidAttributeSyntax() from exc

class NameAndOptionalUIDSyntaxDefinition(StringSyntaxDefinition):
	def encode(self, schema, value):
		return str(value).encode('utf8')

	def decode(self, schema, raw_value):
		try:
			return dn.DNWithUID.from_str(schema, raw_value.decode('utf8'))
		except (UnicodeDecodeError, TypeError, ValueError) as exc:
			raise exceptions.LDAPInvalidAttributeSyntax() from exc

class GeneralizedTimeSyntaxDefinition(SyntaxDefinition):
	def encode(self, schema, value):
		if value.microsecond:
			str_value = value.strftime('%Y%m%d%H%M%S.%f')
		elif value.second:
			str_value = value.strftime('%Y%m%d%H%M%S')
		else:
			str_value = value.strftime('%Y%m%d%H%M')
		if value.tzinfo == datetime.timezone.utc:
			str_value += 'Z'
		elif value.tzinfo is not None:
			delta_seconds = value.tzinfo.utcoffset(value).total_seconds()
			if delta_seconds < 0:
				str_value += '-'
				delta_seconds = -delta_seconds
			else:
				str_value += '+'
			hour = delta_seconds // 3600
			minute = (delta_seconds % 3600) // 60
			str_value += '%02d%02d'%(hour, minute)
		return str_value.encode('ascii')

	def decode(self, schema, raw_value):
		try:
			raw_value = raw_value.decode('utf8')
		except UnicodeDecodeError as exc:
			raise exceptions.LDAPInvalidAttributeSyntax() from exc
		match = re.fullmatch(r'([0-9]{10})(|[0-9]{2}|[0-9]{4})(|[,.][0-9]+)(Z|[+-][0-9]{2}|[+-][0-9]{4})', raw_value)
		if match is None:
			raise exceptions.LDAPInvalidAttributeSyntax()
		main, minute_second, fraction, timezone = match.groups()
		fraction = float('0.' + (fraction[1:] or '0'))
		result = datetime.datetime.strptime(main, '%Y%m%d%H')
		if not minute_second:
			result += datetime.timedelta(hours=fraction)
		if len(minute_second) == 2:
			result += datetime.timedelta(minutes=int(minute_second)+fraction)
		elif len(minute_second) == 4:
			minute = minute_second[:2]
			second = minute_second[2:4]
			result += datetime.timedelta(minutes=int(minute), seconds=int(second)+fraction)
		if timezone == 'Z':
			result = result.replace(tzinfo=datetime.timezone.utc)
		elif timezone:
			sign, hour, minute = timezone[0], timezone[1:3], (timezone[3:5] or '00')
			delta = datetime.timedelta(hours=int(hour), minutes=int(minute))
			if sign == '+':
				result = result.replace(tzinfo=datetime.timezone(delta))
			else:
				result = result.replace(tzinfo=datetime.timezone(-delta))
		return result

class PostalAddressSyntaxDefinition(SyntaxDefinition):
	# 3.3.28.  Postal Address
	#
	# A value of the Postal Address syntax is a sequence of strings of one
	# or more arbitrary UCS characters, which form an address in a physical
	# mail system.
	#
	# The LDAP-specific encoding of a value of this syntax is defined by
	# the following ABNF:
	#
	#
	#   PostalAddress = line *( DOLLAR line )
	#   line          = 1*line-char
	#   line-char     = %x00-23
	#                   / (%x5C "24")  ; escaped "$"
	#                   / %x25-5B
	#                   / (%x5C "5C")  ; escaped "\"
	#                   / %x5D-7F
	#                   / UTFMB
	#
	# Each character string (i.e., <line>) of a postal address value is
	# encoded as a UTF-8 [RFC3629] string, except that "\" and "$"
	# characters, if they occur in the string, are escaped by a "\"
	# character followed by the two hexadecimal digit code for the
	# character.  The <DOLLAR> and <UTFMB> rules are defined in [RFC4512].
	#
	# Many servers limit the postal address to no more than six lines of no
	# more than thirty characters each.
	#
	#   Example:
	#      1234 Main St.$Anytown, CA 12345$USA
	#      \241,000,000 Sweepstakes$PO Box 1000000$Anytown, CA 12345$USA
	#
	# The LDAP definition for the Postal Address syntax is:
	#
	#   ( 1.3.6.1.4.1.1466.115.121.1.41 DESC 'Postal Address' )
	#
	# This syntax corresponds to the PostalAddress ASN.1 type from [X.520];
	# that is
	#
	#   PostalAddress ::= SEQUENCE SIZE(1..ub-postal-line) OF
	#       DirectoryString { ub-postal-string }
	#
	# The values of ub-postal-line and ub-postal-string (both integers) are
	# implementation defined.  Non-normative definitions appear in [X.520].

	# Native values are lists of str
	def encode(self, schema, value):
		return '$'.join([line.replace('\\', '\\5C').replace('$', '\\24') for line in value]).encode('utf8')

	def decode(self, schema, raw_value):
		return [line.replace('\\24', '$').replace('\\5C', '\\') for line in raw_value.decode('utf8').split('$')]

class SubstringAssertionSyntaxDefinition(SyntaxDefinition):
	# Native values are lists of str
	def encode(self, schema, value):
		raise NotImplementedError()

	def decode(self, schema, raw_value):
		value = raw_value.decode('utf8')
		if '*' not in value:
			raise exceptions.LDAPInvalidAttributeSyntax()
		substrings = [substring.replace('\\2A', '*').replace('\\5C', '\\') for substring in value.split('*')]
		initial_substring, *any_substring, final_substring = substrings
		return (initial_substring or None, any_substring, final_substring or None)

class UTCTimeSyntaxDefinition(SyntaxDefinition):
	def encode(self, schema, value):
		if value.second:
			str_value = value.strftime('%y%m%d%H%M%S')
		else:
			str_value = value.strftime('%y%m%d%H%M')
		if value.tzinfo == datetime.timezone.utc:
			str_value += 'Z'
		elif value.tzinfo is not None:
			delta_seconds = value.tzinfo.utcoffset(value).total_seconds()
			if delta_seconds < 0:
				str_value += '-'
				delta_seconds = -delta_seconds
			else:
				str_value += '+'
			hour = delta_seconds // 3600
			minute = (delta_seconds % 3600) // 60
			str_value += '%02d%02d'%(hour, minute)
		return str_value.encode('ascii')

	def decode(self, schema, raw_value):
		try:
			raw_value = raw_value.decode('utf8')
		except UnicodeDecodeError as exc:
			raise exceptions.LDAPInvalidAttributeSyntax() from exc
		match = re.fullmatch(r'([0-9]{10})(|[0-9]{2})(|Z|[+-][0-9]{4})', raw_value)
		if match is None:
			raise exceptions.LDAPInvalidAttributeSyntax()
		main, seconds, timezone = match.groups()
		result = datetime.datetime.strptime(main, '%y%m%d%H%M')
		if seconds:
			result = result.replace(second=int(seconds))
		if timezone == 'Z':
			result = result.replace(tzinfo=datetime.timezone.utc)
		elif timezone:
			sign, hour, minute = timezone[0], timezone[1:3], timezone[3:5]
			delta = datetime.timedelta(hours=int(hour), minutes=int(minute))
			if sign == '+':
				result = result.replace(tzinfo=datetime.timezone(delta))
			else:
				result = result.replace(tzinfo=datetime.timezone(-delta))
		return result

class StubSyntaxDefinition(SyntaxDefinition):
	def encode(self, schema, value):
		raise NotImplementedError()

# RFC2252 (deprecated legacy syntaxes required for some schemas)
Binary =  BytesSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.5', desc='Binary')

# RFC4517
Fax = BytesSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.23', desc='Fax')
JPEG = BytesSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.28', desc='JPEG')
OctetString = BytesSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.40', desc='Octet String')
DirectoryString = StringSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.15', desc='Directory String', re_pattern='.+')
IA5String = StringSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.26', desc='IA5 String', encoding='ascii', extra_compatability_tags=DirectoryString.compatability_tags)
PrintableString = StringSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.44', desc='Printable String', encoding='ascii', re_pattern='[A-Za-z0-9\'()+,.=/:? -]*', extra_compatability_tags=IA5String.compatability_tags)
CountryString = StringSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.11', desc='Country String', encoding='ascii', re_pattern='[A-Za-z0-9\'()+,.=/:? -]{2}', extra_compatability_tags=PrintableString.compatability_tags)
TelephoneNumber = StringSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.50', desc='Telephone Number', encoding='ascii', re_pattern='[A-Za-z0-9\'()+,.=/:? -]*', extra_compatability_tags=PrintableString.compatability_tags)
OID = StringSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.38', desc='OID', encoding='ascii', re_pattern=r'[0-9]+(\.[0-9])*|[A-Za-z][A-Za-z0-9-]*')
BitString = StringSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.6', desc='Bit String', encoding='ascii', re_pattern='\'[01]*\'B')
DeliveryMethod = StringSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.14', desc='Delivery Method')
FacsimileTelephoneNumber = StringSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.22', desc='Facsimile Telephone Number')
EnhancedGuide = StringSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.21', desc='Enhanced Guide')
Guide = StringSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.25', desc='Guide')
NumericString = StringSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.36', desc='Numeric String', encoding='ascii', re_pattern='[0-9 ]+')
OtherMailbox = StringSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.39', desc='Other Mailbox')
TeletexTerminalIdentifier = StringSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.51', desc='Teletex Terminal Identifier')
TelexNumber = StringSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.52', desc='Telex Number')
INTEGER = IntegerSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.27', desc='INTEGER')
AttributeTypeDescription = SchemaElementSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.3', desc='Attribute Type Description', extra_compatability_tags=['FirstComponent:'+OID.oid])
LDAPSyntaxDescription = SchemaElementSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.54', desc='LDAP Syntax Description', extra_compatability_tags=['FirstComponent:'+OID.oid])
MatchingRuleDescription = SchemaElementSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.30', desc='Matching Rule Description', extra_compatability_tags=['FirstComponent:'+OID.oid])
MatchingRuleUseDescription = SchemaElementSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.31', desc='Matching Rule Use Description', extra_compatability_tags=['FirstComponent:'+OID.oid])
ObjectClassDescription = SchemaElementSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.37', desc='Object Class Description', extra_compatability_tags=['FirstComponent:'+OID.oid])
Boolean = BooleanSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.7', desc='Boolean')
DN = DNSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.12', desc='DN')
NameAndOptionalUID = NameAndOptionalUIDSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.34', desc='Name And Optional UID')
GeneralizedTime = GeneralizedTimeSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.24', desc='Generalized Time')
PostalAddress = PostalAddressSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.41', desc='Postal Address')
SubstringAssertion = SubstringAssertionSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.58', desc='Substring Assertion')
UTCTime = UTCTimeSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.53', desc='UTC Time')
DITContentRuleDescription = StubSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.16', desc='DIT Content Rule Description', extra_compatability_tags=['FirstComponent:'+OID.oid])
DITStructureRuleDescription = StubSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.17', desc='DIT Structure Rule Description', extra_compatability_tags=['FirstComponent:'+INTEGER.oid])
NameFormDescription = StubSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.35', desc='Name Form Description', extra_compatability_tags=['FirstComponent:'+OID.oid])

# RFC4523
X509Certificate = StubSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.8', desc='X.509 Certificate')
X509CertificateList = StubSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.9', desc='X.509 Certificate List')
X509CertificatePair = StubSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.10', desc='X.509 Certificate Pair')
X509SupportedAlgorithm = StubSyntaxDefinition('1.3.6.1.4.1.1466.115.121.1.49', desc='X.509 Supported Algorithm')
X509CertificateExactAssertion = StubSyntaxDefinition('1.3.6.1.1.15.1', desc='X.509 Certificate Exact Assertion')
X509CertificateAssertion = StubSyntaxDefinition('1.3.6.1.1.15.2', desc='X.509 Certificate Assertion')
X509CertificatePairExactAssertion = StubSyntaxDefinition('1.3.6.1.1.15.3', desc='X.509 Certificate Pair Exact Assertion')
X509CertificatePairAssertion = StubSyntaxDefinition('1.3.6.1.1.15.4', desc='X.509 Certificate Pair Assertion')
X509CertificateListExactAssertion = StubSyntaxDefinition('1.3.6.1.1.15.5', desc='X.509 Certificate List Exact Assertion')
X509CertificateListAssertion = StubSyntaxDefinition('1.3.6.1.1.15.6', desc='X.509 Certificate List Assertion')
X509AlgorithmIdentifier = StubSyntaxDefinition('1.3.6.1.1.15.7', desc='X.509 Algorithm Identifier')

# RFC3112
AuthPasswordSyntax = StubSyntaxDefinition('1.3.6.1.4.1.4203.1.1.2', desc='authentication password syntax')

ALL = (
	# RFC2252
	Binary,

	# RFC4517
	AttributeTypeDescription,
	BitString,
	Boolean,
	CountryString,
	DeliveryMethod,
	DirectoryString,
	DITContentRuleDescription,
	DITStructureRuleDescription,
	DN,
	EnhancedGuide,
	FacsimileTelephoneNumber,
	Fax,
	GeneralizedTime,
	Guide,
	IA5String,
	INTEGER,
	JPEG,
	LDAPSyntaxDescription,
	MatchingRuleDescription,
	MatchingRuleUseDescription,
	NameAndOptionalUID,
	NameFormDescription,
	NumericString,
	ObjectClassDescription,
	OctetString,
	OID,
	OtherMailbox,
	PostalAddress,
	PrintableString,
	SubstringAssertion,
	TelephoneNumber,
	TeletexTerminalIdentifier,
	TelexNumber,
	UTCTime,

	# RFC4523
	X509Certificate,
	X509CertificateList,
	X509CertificatePair,
	X509SupportedAlgorithm,
	X509CertificateExactAssertion,
	X509CertificateAssertion,
	X509CertificatePairExactAssertion,
	X509CertificatePairAssertion,
	X509CertificateListExactAssertion,
	X509CertificateListAssertion,
	X509AlgorithmIdentifier,

	# RFC3112
	AuthPasswordSyntax,
)



import enum
import re

from .. import exceptions

__all__ = [
	'SyntaxDefinition',
	'MatchingRuleKind',
	'MatchingRuleDefinition',
	'MatchingRuleUseDefinition',
	'AttributeTypeUsage',
	'AttributeTypeDefinition',
	'ObjectClassKind',
	'ObjectClassDefinition',
]

def escape(string):
	return string.replace('\\', '\\5C').replace('\'', '\\27')

def qdescr_to_token(descr):
	return '\''+descr+'\''

def qdescrs_to_tokens(descrs):
	if len(descrs) == 1:
		return [qdescr_to_token(descrs[0])]
	return ['('] + [qdescr_to_token(descr) for descr in descrs] + [')']

def oids_to_tokens(oids):
	if len(oids) == 1:
		return [oids[0]]
	tokens = ['(', oids[0]]
	for oid in oids[1:]:
		tokens += ['$', oid]
	return tokens + [')']

def qdstring_to_token(qdstring):
	return '\''+escape(qdstring)+'\''

def qdstrings_to_tokens(qdstrings):
	if len(qdstrings) == 1:
		return [qdstring_to_token(qdstrings[0])]
	return ['('] + [qdstring_to_token(qdstring) for qdstring in qdstrings] + [')']

def extensions_to_tokens(extensions):
	tokens = []
	if not extensions:
		return []
	for key, values in extensions.items():
		if not key.startswith('X-'):
			raise ValueError('Extention names must start with "X-"')
		tokens += [key] + qdstrings_to_tokens(values)
	return tokens

def tokenize(string):
	tokens = []
	offset = 0
	while string:
		match = re.match(r" *([()$]|[A-Za-z0-9.{}-]+|'[^']*') *", string)
		if not match:
			string_abbrev = string[:20] + '...'
			raise ValueError(f'Unrecognized token at offset {offset}: "{string_abbrev}"')
		tokens.append(match.groups()[0])
		string = string[match.end():]
		offset += match.end()
	return tokens

def pop_token(tokens):
	if not tokens:
		raise ValueError('Unexpected end of input')
	return tokens.pop(0)

def check_token(tokens, expected_token):
	if not tokens or tokens[0] != expected_token:
		return False
	tokens.pop(0)
	return True

def parse_token(tokens, expected_token):
	token = pop_token(tokens)
	if token != expected_token:
		raise ValueError(f'Expected "{expected_token}" but got "{token}"')
	return token

def parse_numericoid(tokens):
	token = pop_token(tokens)
	if not re.fullmatch(r"([0-9]|[1-9][0-9]*)(\.([0-9]|[1-9][0-9]*))*", token):
		raise ValueError(f'Invalid numeric OID "{token}"')
	return token

def parse_qdescr(tokens):
	token = pop_token(tokens)
	match = re.fullmatch(r"'([A-Za-z][A-Za-z0-9-]*)'", token)
	if not match:
		raise ValueError(f'Invalid quoted descriptor "{token}"')
	return match.groups()[0]

def parse_qdescrs(tokens):
	if check_token(tokens, '('):
		result = []
		while not check_token(tokens, ')'):
			result.append(parse_qdescr(tokens))
		return result
	return [parse_qdescr(tokens)]

def parse_qdstring(tokens):
	token = pop_token(tokens)
	match = re.fullmatch(r"'(([^'\\]|\\27|\\5C|\\5c)+)'", token)
	if not match:
		raise ValueError(f'Invalid quoted string "{token}"')
	return match.groups()[0].replace('\\27', '\'').replace('\\5C', '\\').replace('\\5c', '\\')

def parse_qdstrings(tokens):
	if check_token(tokens, '('):
		result = []
		while not check_token(tokens, ')'):
			result.append(parse_qdstring(tokens))
		return result
	return [parse_qdstring(tokens)]

def parse_oid(tokens):
	token = pop_token(tokens)
	if not re.fullmatch(r'([0-9]|[1-9][0-9]*)(\.([0-9]|[1-9][0-9]*))*|[A-Za-z][A-Za-z0-9-]*', token):
		raise ValueError(f'Invalid OID "{token}"')
	return token

def parse_oids(tokens):
	if check_token(tokens, '('):
		result = [parse_oid(tokens)]
		while not check_token(tokens, ')'):
			parse_token(tokens, '$')
			result.append(parse_oid(tokens))
		return result
	return [parse_oid(tokens)]

def parse_noidlen(tokens):
	token = pop_token(tokens)
	match = re.fullmatch(r"(([0-9]|[1-9][0-9]*)(\.([0-9]|[1-9][0-9]*))*)(|\{([0-9]|[1-9][0-9]*)\})", token)
	if not match:
		raise ValueError(f'Invalid numeric OID with optional length "{token}"')
	noid, _, _, _, _, length = match.groups()
	if length is not None:
		length = int(length)
	return noid, length

def parse_extensions(tokens):
	results = {}
	while tokens and re.fullmatch(r"X-[A-Z_-]+", tokens[0]):
		name = tokens.pop(0)
		values = parse_qdstrings(tokens)
		results[name] = values
	return results

class SyntaxDefinition:
	def __init__(self, oid, desc='', extensions=None, extra_compatability_tags=None):
		self.oid = oid
		self.desc = desc
		self.extensions = extensions or {}
		self.compatability_tags = {oid} | set(extra_compatability_tags or tuple())

	def __str__(self):
		tokens = ['(', self.oid]
		if self.desc:
			tokens += ['DESC', qdstring_to_token(self.desc)]
		tokens += extensions_to_tokens(self.extensions) + [')']
		return ' '.join(tokens)

	@property
	def first_component_oid(self):
		return self.oid

	def encode(self, schema, value):
		'''Encode native value to its LDAP-specific encoding

		:param schema: Schema of the object in whose context encoding takes place
		:type schema: Schema
		:param value: native value (depends on syntax)
		:type value: any

		:returns: LDAP-specific encoding of the value
		:rtype: bytes'''
		raise NotImplementedError()

	def decode(self, schema, raw_value):
		'''Decode LDAP-specific encoding of a value to a native value

		:param schema: Schema of the object in whose context decoding takes place
		:type schema: Schema
		:param raw_value: LDAP-specific encoding of the value
		:type raw_value: bytes

		:returns: native value (depends on syntax)
		:rtype: any
		:raises exceptions.LDAPError: if raw_value is invalid'''
		raise exceptions.LDAPInvalidAttributeSyntax()

class MatchingRuleKind(enum.Enum):
	EQUALITY = enum.auto()
	ORDERING = enum.auto()
	SUBSTR = enum.auto()

class MatchingRuleDefinition:
	# pylint: disable=too-many-arguments
	def __init__(self, oid, name=None, desc='', obsolete=False, syntax=None, extensions=None, compatability_tag=None, kind=None):
		if not syntax:
			raise ValueError('syntax must be specified')
		if not kind:
			raise ValueError('kind must be specified')
		self.oid = oid
		self.syntax = syntax
		self.name = name or []
		self.desc = desc
		self.obsolete = obsolete
		self.extensions = extensions or {}
		self.compatability_tag = compatability_tag or syntax
		self.kind = kind

	def __str__(self):
		tokens = ['(', self.oid]
		if self.name:
			tokens += ['NAME'] + qdescrs_to_tokens(self.name)
		if self.desc:
			tokens += ['DESC', qdstring_to_token(self.desc)]
		if self.obsolete:
			tokens += ['OBSOLETE']
		tokens += ['SYNTAX', self.syntax]
		tokens += extensions_to_tokens(self.extensions) + [')']
		return ' '.join(tokens)

	@property
	def first_component_oid(self):
		return self.oid

	def match_equal(self, schema, attribute_values, assertion_value):
		'''Return whether any attribute value equals assertion value

		Only available for EQUALITY matching rules.

		:returns: True if any attribute values matches, False otherwise
		:rtype: bool
		:raises exceptions.LDAPError: if the result is undefined'''
		raise exceptions.LDAPInappropriateMatching()

	def match_approx(self, schema, attribute_values, assertion_value):
		'''Return whether any attribute value approximatly equals assertion value

		Only available for EQUALITY matching rules.

		:returns: True if any attribute values matches, False otherwise
		:rtype: bool
		:raises exceptions.LDAPError: if the result is undefined'''
		return self.match_equal(schema, attribute_values, assertion_value)

	def match_less(self, schema, attribute_values, assertion_value):
		'''Return whether any attribute value is less than assertion value

		Only available for ORDERING matching rules.

		:returns: True if any attribute values matches, False otherwise
		:rtype: bool
		:raises exceptions.LDAPError: if the result is undefined'''
		raise exceptions.LDAPInappropriateMatching()

	def match_greater_or_equal(self, schema, attribute_values, assertion_value):
		'''Return whether any attribute value is greater than or equal to assertion value

		Only available for ORDERING matching rules.

		:returns: True if any attribute values matches, False otherwise
		:rtype: bool
		:raises exceptions.LDAPError: if the result is undefined'''
		raise exceptions.LDAPInappropriateMatching()

	def match_substr(self, schema, attribute_values, inital_substring, any_substrings, final_substring):
		'''Return whether any attribute value matches a substring assertion

		Only available for SUBSTR matching rules.

		The type of `inital_substring`, `any_substrings` and `final_substring`
		depends on the syntax of the attribute's equality matching rule!

		:param schema: Schema of the object whose attribute values are matched
		:type schema: Schema
		:param attribute_values: Attribute values (type according to attribute's syntax)
		:type attribute_values: List of any
		:param inital_substring: Substring to match the beginning (optional)
		:type inital_substring: any
		:param any_substrings: List of substrings to match between initial and final in order
		:type any_substrings: list of any
		:param final_substring: Substring to match the end (optional)
		:type final_substring: any
		:returns: True if any attribute values matches, False otherwise
		:rtype: bool
		:raises exceptions.LDAPError: if the result is undefined'''
		raise exceptions.LDAPInappropriateMatching()

class MatchingRuleUseDefinition:
	def __init__(self, oid, name=None, desc='', obsolete=False, applies=None, extensions=None):
		self.oid = oid
		self.name = name or []
		self.desc = desc
		self.obsolete = obsolete
		self.applies = applies or []
		self.extensions = extensions or {}

	def __str__(self):
		tokens = ['(', self.oid]
		if self.name:
			tokens += ['NAME'] + qdescrs_to_tokens(self.name)
		if self.desc:
			tokens += ['DESC', qdstring_to_token(self.desc)]
		if self.obsolete:
			tokens += ['OBSOLETE']
		tokens += ['APPLIES'] + oids_to_tokens(self.applies)
		tokens += extensions_to_tokens(self.extensions) + [')']
		return ' '.join(tokens)

	@property
	def first_component_oid(self):
		return self.oid


class AttributeTypeUsage(enum.Enum):
	'''Values for usage argument of `AttributeTypeUsage`'''
	# pylint: disable=invalid-name
	# user
	userApplications = enum.auto()
	# directory operational
	directoryOperation = enum.auto()
	# DSA-shared operational
	distributedOperation = enum.auto()
	# DSA-specific operational
	dSAOperation = enum.auto()

class AttributeTypeDefinition:
	# pylint: disable=too-many-arguments,too-many-instance-attributes,too-many-locals
	def __init__(self, oid, name=None, desc='', obsolete=False, sup=None,
	             equality=None, ordering=None, substr=None, syntax=None,
               syntax_len=None, single_value=False, collective=False,
		           no_user_modification=False,
	             usage=AttributeTypeUsage.userApplications, extensions=None):
		if not sup and not syntax:
			raise ValueError('Either SUP or SYNTAX must be specified')
		self.oid = oid
		self.name = name or []
		self.desc = desc
		self.obsolete = obsolete
		self.sup = sup
		self.equality = equality
		self.ordering = ordering
		self.substr = substr
		self.syntax = syntax
		self.syntax_len = syntax_len
		self.single_value = single_value
		self.collective = collective
		self.no_user_modification = no_user_modification
		self.usage = usage
		self.extensions = extensions or {}

	def __str__(self):
		tokens = ['(', self.oid]
		if self.name:
			tokens += ['NAME'] + qdescrs_to_tokens(self.name)
		if self.desc:
			tokens += ['DESC', qdstring_to_token(self.desc)]
		if self.obsolete:
			tokens += ['OBSOLETE']
		if self.sup:
			tokens += ['SUP', self.sup]
		if self.equality:
			tokens += ['EQUALITY', self.equality]
		if self.ordering:
			tokens += ['ORDERING', self.ordering]
		if self.substr:
			tokens += ['SUBSTR', self.substr]
		if self.syntax:
			if self.syntax_len is None:
				tokens += ['SYNTAX', self.syntax]
			else:
				tokens += ['SYNTAX', self.syntax+'{'+str(self.syntax_len)+'}']
		if self.single_value:
			tokens += ['SINGLE-VALUE']
		if self.collective:
			tokens += ['COLLECTIVE']
		if self.no_user_modification:
			tokens += ['NO-USER-MODIFICATION']
		if self.usage != AttributeTypeUsage.userApplications:
			tokens += ['USAGE', self.usage.name]
		tokens += extensions_to_tokens(self.extensions) + [')']
		return ' '.join(tokens)

	@classmethod
	def from_str(cls, string):
		tokens = tokenize(string)
		parse_token(tokens, '(')
		oid = parse_numericoid(tokens)
		name = []
		if check_token(tokens, 'NAME'):
			name = parse_qdescrs(tokens)
		desc = ''
		if check_token(tokens, 'DESC'):
			desc = parse_qdstring(tokens)
		obsolete = check_token(tokens, 'OBSOLETE')
		sup = None
		if check_token(tokens, 'SUP'):
			sup = parse_oid(tokens)
		equality = None
		if check_token(tokens, 'EQUALITY'):
			equality = parse_oid(tokens)
		ordering = None
		if check_token(tokens, 'ORDERING'):
			ordering = parse_oid(tokens)
		substr = None
		if check_token(tokens, 'SUBSTR'):
			substr = parse_oid(tokens)
		syntax, syntax_len = None, None
		if check_token(tokens, 'SYNTAX'):
			syntax, syntax_len = parse_noidlen(tokens)
		single_value = check_token(tokens, 'SINGLE-VALUE')
		collective = check_token(tokens, 'COLLECTIVE')
		no_user_modification = check_token(tokens, 'NO-USER-MODIFICATION')
		usage = AttributeTypeUsage.userApplications
		if check_token(tokens, 'USAGE'):
			token = tokens.pop(0)
			if token == 'userApplications':
				usage = AttributeTypeUsage.userApplications
			elif token == 'directoryOperation':
				usage = AttributeTypeUsage.directoryOperation
			elif token == 'distributedOperation':
				usage = AttributeTypeUsage.distributedOperation
			elif token == 'dSAOperation':
				usage = AttributeTypeUsage.dSAOperation
			else:
				raise ValueError(f'Invalid usage value "{token}"')
		extensions = parse_extensions(tokens)
		parse_token(tokens, ')')
		if tokens:
			raise ValueError(f'Unexpected token "{tokens[0]}", expected no more input')
		return cls(oid, name=name, desc=desc, obsolete=obsolete, sup=sup,
		           equality=equality, ordering=ordering, substr=substr,
		           syntax=syntax, syntax_len=syntax_len, single_value=single_value,
		           collective=collective, no_user_modification=no_user_modification,
		           usage=usage, extensions=extensions)

	@property
	def first_component_oid(self):
		return self.oid

class ObjectClassKind(enum.Enum):
	'''Values for kind argument of `ObjectClass`'''
	ABSTRACT = enum.auto()
	STRUCTURAL = enum.auto()
	AUXILIARY = enum.auto()

class ObjectClassDefinition:
	# pylint: disable=too-many-arguments
	def __init__(self, oid, name=None, desc='', obsolete=False, sup=None,
	             kind=ObjectClassKind.STRUCTURAL, must=None, may=None,
	             extensions=None):
		self.oid = oid
		self.name = name or []
		self.desc = desc
		self.obsolete = obsolete
		self.sup = sup or []
		self.kind = kind
		self.must = must or []
		self.may = may or []
		self.extensions = extensions or {}

	def __str__(self):
		tokens = ['(', self.oid]
		if self.name:
			tokens += ['NAME'] + qdescrs_to_tokens(self.name)
		if self.desc:
			tokens += ['DESC', qdstring_to_token(self.desc)]
		if self.obsolete:
			tokens += ['OBSOLETE']
		if self.sup:
			tokens += ['SUP'] + oids_to_tokens(self.sup)
		tokens += [self.kind.name]
		if self.must:
			tokens += ['MUST'] + oids_to_tokens(self.must)
		if self.may:
			tokens += ['MAY'] + oids_to_tokens(self.may)
		tokens += extensions_to_tokens(self.extensions) + [')']
		return ' '.join(tokens)

	@classmethod
	def from_str(cls, string):
		tokens = tokenize(string)
		parse_token(tokens, '(')
		oid = parse_numericoid(tokens)
		name = []
		if check_token(tokens, 'NAME'):
			name = parse_qdescrs(tokens)
		desc = ''
		if check_token(tokens, 'DESC'):
			desc = parse_qdstring(tokens)
		obsolete = check_token(tokens, 'OBSOLETE')
		sup = []
		if check_token(tokens, 'SUP'):
			sup = parse_oids(tokens)
		kind = ObjectClassKind.STRUCTURAL
		if check_token(tokens, 'ABSTRACT'):
			kind = ObjectClassKind.ABSTRACT
		elif check_token(tokens, 'STRUCTURAL'):
			kind = ObjectClassKind.STRUCTURAL
		elif check_token(tokens, 'AUXILIARY'):
			kind = ObjectClassKind.AUXILIARY
		must = []
		if check_token(tokens, 'MUST'):
			must = parse_oids(tokens)
		may = []
		if check_token(tokens, 'MAY'):
			may = parse_oids(tokens)
		extensions = parse_extensions(tokens)
		parse_token(tokens, ')')
		if tokens:
			raise ValueError(f'Unexpected token "{tokens[0]}", expected no more input')
		return cls(oid, name=name, desc=desc, obsolete=obsolete, sup=sup,
		           kind=kind, must=must, may=may, extensions=extensions)

	@property
	def first_component_oid(self):
		return self.oid
